"""
The VOMS2CSAgent extracts VO user information from VOMS servers.

It performs the following operations:

- Extracts user info from the VOMS server using its REST interface
- Finds user DN's not yet registered in the DIRAC Registry
- For each new DN it constructs a DIRAC login name by a best guess or using the nickname VOMS attribute
- Registers new users to the DIRAC Registry including group membership
- Updates information for already registered users, including group membership
- Sends report for performed operation to the VO administrator

The agent is performing its operations with credentials of the VO administrator as defined
in the ``/Registry/VO/<VO_name>`` configuration section.

The following options can be set for the VOMS2CSAgent.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN VOMS2CSAgent
  :end-before: ##END
  :dedent: 2
  :caption: VOMS2CSAgent options

.. note:: The options *AutoAddUsers*, *AutoModifyUsers*, *AutoDeleteUsers* can be overridden by the
  corresponding options defined in the ``/Registry/VO/<VO_name>`` configuration section.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK, gConfig, S_ERROR
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.Core.Utilities.Proxy import executeWithUserProxy
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOOption, getUserOption
from DIRAC.ConfigurationSystem.Client.VOMS2CSSynchronizer import VOMS2CSSynchronizer
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog

__RCSID__ = "$Id$"


class VOMS2CSAgent(AgentModule):
    def __init__(self, *args, **kwargs):
        """Defines default parameters"""
        super(VOMS2CSAgent, self).__init__(*args, **kwargs)

        self.voList = ["any"]
        self.dryRun = True

        self.autoAddUsers = True
        self.autoModifyUsers = True
        self.autoDeleteUsers = True
        self.detailedReport = True
        self.makeFCEntry = False
        self.autoLiftSuspendedStatus = True
        self.mailFrom = "noreply@dirac.system"
        self.syncPluginName = None

    def initialize(self):
        """Initialize the default parameters"""

        self.dryRun = self.am_getOption("DryRun", self.dryRun)

        # # General agent options, can be overridden by VO options
        self.autoAddUsers = self.am_getOption("AutoAddUsers", self.autoAddUsers)
        self.autoModifyUsers = self.am_getOption("AutoModifyUsers", self.autoModifyUsers)
        self.autoDeleteUsers = self.am_getOption("AutoDeleteUsers", self.autoDeleteUsers)
        self.autoLiftSuspendedStatus = self.am_getOption("AutoLiftSuspendedStatus", self.autoLiftSuspendedStatus)
        self.makeFCEntry = self.am_getOption("MakeHomeDirectory", self.makeFCEntry)
        self.syncPluginName = self.am_getOption("SyncPluginName", self.syncPluginName)

        self.detailedReport = self.am_getOption("DetailedReport", self.detailedReport)
        self.mailFrom = self.am_getOption("MailFrom", self.mailFrom)

        self.voList = self.am_getOption("VO", self.voList)
        if not self.voList:
            return S_ERROR("Option 'VO' not configured")
        if self.voList[0].lower() == "any":
            result = gConfig.getSections("/Registry/VO")
            if not result["OK"]:
                return result
            self.voList = result["Value"]
            self.log.notice("VOs", self.voList)

        return S_OK()

    def execute(self):

        for vo in self.voList:
            voAdminUser = getVOOption(vo, "VOAdmin")
            voAdminMail = None
            if voAdminUser:
                voAdminMail = getUserOption(voAdminUser, "Email")
            voAdminGroup = getVOOption(vo, "VOAdminGroup", getVOOption(vo, "DefaultGroup"))

            self.log.info("Performing VOMS sync", "for VO %s with credentials %s@%s" % (vo, voAdminUser, voAdminGroup))

            autoAddUsers = getVOOption(vo, "AutoAddUsers", self.autoAddUsers)
            autoModifyUsers = getVOOption(vo, "AutoModifyUsers", self.autoModifyUsers)
            autoDeleteUsers = getVOOption(vo, "AutoDeleteUsers", self.autoDeleteUsers)
            autoLiftSuspendedStatus = getVOOption(vo, "AutoLiftSuspendedStatus", self.autoLiftSuspendedStatus)
            syncPluginName = getVOOption(vo, "SyncPluginName", self.syncPluginName)

            vomsSync = VOMS2CSSynchronizer(
                vo,
                autoAddUsers=autoAddUsers,
                autoModifyUsers=autoModifyUsers,
                autoDeleteUsers=autoDeleteUsers,
                autoLiftSuspendedStatus=autoLiftSuspendedStatus,
                syncPluginName=syncPluginName,
            )

            result = self.__syncCSWithVOMS(  # pylint: disable=unexpected-keyword-arg
                vomsSync,
                proxyUserName=voAdminUser,
                proxyUserGroup=voAdminGroup,
            )
            if not result["OK"]:
                self.log.error("Failed to perform VOMS to CS synchronization:", "VO %s: %s" % (vo, result["Message"]))
                continue
            resultDict = result["Value"]
            newUsers = resultDict.get("NewUsers", [])
            modUsers = resultDict.get("ModifiedUsers", [])
            delUsers = resultDict.get("DeletedUsers", [])
            susUsers = resultDict.get("SuspendedUsers", [])
            csapi = resultDict.get("CSAPI")
            adminMessages = resultDict.get("AdminMessages", {"Errors": [], "Info": []})
            voChanged = resultDict.get("VOChanged", False)
            self.log.info(
                "Run user results",
                ": new %d, modified %d, deleted %d, new/suspended %d"
                % (len(newUsers), len(modUsers), len(delUsers), len(susUsers)),
            )

            if csapi.csModified:
                # We have accumulated all the changes, commit them now
                self.log.info("There are changes to the CS ready to be committed", "for VO %s" % vo)
                if self.dryRun:
                    self.log.info("Dry Run: CS won't be updated")
                    csapi.showDiff()
                else:
                    result = csapi.commitChanges()
                    if not result["OK"]:
                        self.log.error("Could not commit configuration changes", result["Message"])
                        return result
                    self.log.notice("Configuration committed", "for VO %s" % vo)
            else:
                self.log.info("No changes to the CS recorded at this cycle", "for VO %s" % vo)

            # Add user home directory in the file catalog
            if self.makeFCEntry and newUsers:
                self.log.info("Creating home directories for users", str(newUsers))
                result = self.__addHomeDirectory(  # pylint: disable=unexpected-keyword-arg
                    vo,
                    newUsers,
                    proxyUserName=voAdminUser,
                    proxyUserGroup=voAdminGroup,
                )
                if not result["OK"]:
                    self.log.error("Failed to create user home directories:", "VO %s: %s" % (vo, result["Message"]))
                else:
                    for user in result["Value"]["Failed"]:
                        self.log.error(
                            "Failed to create home directory",
                            "user: %s, operation: %s" % (user, result["Value"]["Failed"][user]),
                        )
                        adminMessages["Errors"].append(
                            "Failed to create home directory for user %s: operation %s"
                            % (user, result["Value"]["Failed"][user])
                        )
                    for user in result["Value"]["Successful"]:
                        adminMessages["Info"].append("Created home directory for user %s" % user)

            if voChanged or self.detailedReport:
                mailMsg = ""
                if adminMessages["Errors"]:
                    mailMsg += "\nErrors list:\n  %s" % "\n  ".join(adminMessages["Errors"])
                if adminMessages["Info"]:
                    mailMsg += "\nRun result:\n  %s" % "\n  ".join(adminMessages["Info"])
                if self.detailedReport:
                    result = vomsSync.getVOUserReport()
                    if result["OK"]:
                        mailMsg += "\n\n"
                        mailMsg += result["Value"]
                    else:
                        mailMsg += "Failed to produce a detailed user report"
                        mailMsg += result["Message"]
                if self.dryRun:
                    self.log.info("Dry Run: mail won't be sent")
                    self.log.info(mailMsg)
                else:
                    NotificationClient().sendMail(
                        self.am_getOption("MailTo", voAdminMail), "VOMS2CSAgent run log", mailMsg, self.mailFrom
                    )

        return S_OK()

    @executeWithUserProxy
    def __syncCSWithVOMS(self, vomsSync):
        return vomsSync.syncCSWithVOMS()

    @executeWithUserProxy
    def __addHomeDirectory(self, vo, newUsers):

        fc = FileCatalog(vo=vo)
        defaultVOGroup = getVOOption(vo, "DefaultGroup", "%s_user" % vo)

        failed = {}
        successful = {}
        for user in newUsers:
            result = fc.addUser(user)
            if not result["OK"]:
                failed[user] = "addUser"
                continue
            dirName = "/%s/user/%s/%s" % (vo, user[0], user)
            result = fc.createDirectory(dirName)
            if not result["OK"]:
                failed[user] = "createDirectory"
                continue
            result = fc.changePathOwner({dirName: user}, recursive=False)
            if not result["OK"]:
                failed[user] = "changePathOwner"
                continue
            result = fc.changePathGroup({dirName: defaultVOGroup}, recursive=False)
            if not result["OK"]:
                failed[user] = "changePathGroup"
                continue
            successful[user] = True

        return S_OK({"Successful": successful, "Failed": failed})



"""
Accounting agent to consume perfSONAR network metrics received via a message queue.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN NetworkAgent
  :end-before: ##END
  :dedent: 2
  :caption: NetworkAgent options

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from datetime import datetime

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.AccountingSystem.Client.Types.Network import Network
from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC.Resources.MessageQueue.MQCommunication import createConsumer

__RCSID__ = "$Id$"


class NetworkAgent(AgentModule):
    """
    AccountingSystem agent to processes messages containing perfSONAR network metrics.
    Results are stored in the accounting database.
    """

    #: default number of seconds after which network accounting
    #: objects are removed from the temporary buffer
    BUFFER_TIMEOUT = 3600

    def initialize(self):

        self.log = gLogger.getSubLogger("NetworkAgent")

        # API initialization is required to get an up-to-date configuration from the CS
        self.csAPI = CSAPI()
        self.csAPI.initialize()

        # temporary buffer for network accounting objects + some parameters
        self.buffer = {}  # { {addTime: datetime.now(), object: Network() }, ... }
        self.bufferTimeout = self.am_getOption("BufferTimeout", NetworkAgent.BUFFER_TIMEOUT)

        # internal list of message queue consumers
        self.consumers = []

        # host-to-dirac name dictionary
        self.nameDictionary = {}

        # statistics
        self.messagesCount = 0  # number of received messages
        self.messagesCountOld = 0  # previous number of received messages (used to check connection status)

        self.skippedMessagesCount = 0  # number of skipped messages (errors, unsupported metrics, etc.)
        self.PLRMetricCount = 0  # number of received packet-loss-rate metrics
        self.OWDMetricCount = 0  # number of received one-way-delay metrics
        self.skippedMetricCount = 0  # number of skipped metrics (errors, invalid data, etc.)
        self.insertedCount = 0  # number of properly inserted accounting objects
        self.removedCount = 0  # number of removed accounting objects (missing data)

        return S_OK()

    def finalize(self):
        """
        Gracefully close all consumer connections and commit last records to the DB.
        """

        for consumer in self.consumers:
            consumer.close()

        self.commitData()

        return S_OK()

    def execute(self):
        """
        During each cycle update the internal host-to-dirac name dictionary,
        check the consumers status (restart them if necessary),
        commit data stored in the buffer and show statistics.
        """

        self.updateNameDictionary()
        self.checkConsumers()
        self.commitData()
        self.showStatistics()

        return S_OK()

    def updateNameDictionary(self):
        """
        Update the internal host-to-dirac name dictionary.
        """

        result = gConfig.getConfigurationTree("/Resources/Sites", "Network/", "/Enabled")
        if not result["OK"]:
            self.log.error("getConfigurationTree() failed with message: %s" % result["Message"])
            return S_ERROR("Unable to fetch perfSONAR endpoints from CS.")

        tmpDict = {}
        for path, value in result["Value"].items():
            if value == "True":
                elements = path.split("/")
                diracName = elements[4]
                hostName = elements[6]
                tmpDict[hostName] = diracName

        self.nameDictionary = tmpDict

    def checkConsumers(self):
        """
        Check whether consumers exist and work properly.
        (Re)create consumers if needed.
        """

        # recreate consumers if there are any problems
        if not self.consumers or self.messagesCount == self.messagesCountOld:

            for consumer in self.consumers:
                consumer.close()

            for uri in self.am_getOption("MessageQueueURI", "").replace(" ", "").split(","):
                result = createConsumer(uri, self.processMessage)
                if not result["OK"]:
                    self.log.error("Failed to create a consumer from URI: %s" % uri)
                    continue
                else:
                    self.log.info("Successfully created a consumer from URI: %s" % uri)

                self.consumers.append(result["Value"])

            if self.consumers:
                return S_OK("Successfully created at least one consumer")
            return S_ERROR("Failed to create at least one consumer")

        # if everything is OK just update the counter
        else:
            self.messagesCountOld = self.messagesCount

    def processMessage(self, headers, body):
        """
        Process a message containing perfSONAR data and store the result in the Accounting DB.
        Supports packet-loss-rate and one-way-delay metrics send in raw data streams.

        Function is designed to be an MQConsumer callback function.
        """

        self.messagesCount += 1
        metadata = {
            "SourceIP": body["meta"]["source"],
            "SourceHostName": body["meta"]["input_source"],
            "DestinationIP": body["meta"]["destination"],
            "DestinationHostName": body["meta"]["input_destination"],
        }

        try:
            metadata["Source"] = self.nameDictionary[body["meta"]["input_source"]]
            metadata["Destination"] = self.nameDictionary[body["meta"]["input_destination"]]
        except KeyError as error:
            # messages with unsupported source or destination host name can be safely skipped
            self.skippedMessagesCount += 1
            self.log.debug('Host "%s" does not exist in the host-to-dirac name dictionary (message skipped)' % error)
            return S_OK()

        metadataKey = ""
        for value in metadata.values():
            metadataKey += value

        timestamps = sorted(body["datapoints"])
        for timestamp in timestamps:
            try:
                date = datetime.utcfromtimestamp(float(timestamp))

                # create a key that allows to join packet-loss-rate and one-way-delay
                # metrics in one network accounting record
                networkAccountingObjectKey = "%s%s" % (metadataKey, str(date))

                # use existing or create a new temporary accounting
                # object to store the data in DB
                if networkAccountingObjectKey in self.buffer:
                    net = self.buffer[networkAccountingObjectKey]["object"]

                    timeDifference = datetime.now() - self.buffer[networkAccountingObjectKey]["addTime"]
                    if timeDifference.total_seconds() > 60:
                        self.log.warn("Object was taken from buffer after %s" % (timeDifference))
                else:
                    net = Network()
                    net.setStartTime(date)
                    net.setEndTime(date)
                    net.setValuesFromDict(metadata)

                # get data stored in metric
                metricData = body["datapoints"][timestamp]

                # look for supported event types
                if headers["event-type"] == "packet-loss-rate":
                    self.PLRMetricCount += 1
                    if metricData < 0 or metricData > 1:
                        raise Exception("Invalid PLR metric (%s)" % (metricData))

                    net.setValueByKey("PacketLossRate", metricData * 100)
                elif headers["event-type"] == "histogram-owdelay":
                    self.OWDMetricCount += 1

                    # calculate statistics from histogram
                    OWDMin = 999999
                    OWDMax = 0
                    total = 0
                    count = 0
                    for value, items in metricData.items():
                        floatValue = float(value)
                        total += floatValue * items
                        count += items
                        OWDMin = min(OWDMin, floatValue)
                        OWDMax = max(OWDMax, floatValue)
                    OWDAvg = float(total) / count

                    # skip metrics with invalid data
                    if OWDAvg < 0 or OWDMin < 0 or OWDMax < 0:
                        raise Exception("Invalid OWD metric (%s, %s, %s)" % (OWDMin, OWDAvg, OWDMax))
                    else:
                        # approximate jitter value
                        net.setValueByKey("Jitter", OWDMax - OWDMin)
                        net.setValueByKey("OneWayDelay", OWDAvg)

                else:
                    self.skippedMetricCount += 1
                    continue

                self.buffer[networkAccountingObjectKey] = {"addTime": datetime.now(), "object": net}

            # suppress all exceptions to protect the listener thread
            except Exception as e:
                self.skippedMetricCount += 1
                self.log.warn("Metric skipped because of an exception: %s" % e)

        return S_OK()

    def commitData(self):
        """
        Iterates through all object in the temporary buffer and commit objects to DB
        if both packet-loss-rate and one-way-delay values are set.

        Objects in the buffer older than self.bufferTimeout seconds which still have
        missing data are removed permanently (a warning is issued).
        """

        now = datetime.now()
        removed = False

        for key, value in self.buffer.items():
            result = value["object"].checkValues()
            if not result["OK"]:
                if (now - value["addTime"]).total_seconds() > self.bufferTimeout:
                    del self.buffer[key]
                    self.removedCount += 1
                    removed = True
            else:
                value["object"].delayedCommit()
                del self.buffer[key]
                self.insertedCount += 1

        if removed:
            self.log.warn("Network accounting object(s) has been removed because of missing data")

        return S_OK()

    def showStatistics(self):
        """Display different statistics as info messages in the log file."""

        self.log.info("\tReceived messages:           %s" % self.messagesCount)
        self.log.info("\tSkipped messages:            %s" % self.skippedMessagesCount)
        self.log.info("\tPacket-Loss-Rate metrics:    %s" % self.PLRMetricCount)
        self.log.info("\tOne-Way-Delay metrics:       %s" % self.OWDMetricCount)
        self.log.info("\tSkipped metrics:             %s" % self.skippedMetricCount)
        self.log.info("")
        self.log.info("\tObjects in the buffer:       %s" % len(self.buffer))
        self.log.info("\tObjects inserted to DB:      %s" % self.insertedCount)
        self.log.info("\tPermanently removed objects: %s" % self.removedCount)

        return S_OK()



""" py.test test of ExecutorDispatcher
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access

__RCSID__ = "$Id$"


from DIRAC.Core.Utilities.ExecutorDispatcher import (
    ExecutorState,
    ExecutorQueues,
)


execState = ExecutorState()
execState.addExecutor(1, "type1", 2)

eQ = ExecutorQueues()


def test_ExecutorState():
    """test of ExecutorState"""
    assert execState.freeSlots(1) == 2
    assert execState.addTask(1, "t1") == 1
    assert execState.addTask(1, "t1") == 1
    assert execState.addTask(1, "t2") == 2
    assert execState.freeSlots(1) == 0
    assert execState.full(1)
    assert execState.removeTask("t1") == 1
    assert execState.freeSlots(1) == 1
    assert execState.getFreeExecutors("type1") == {1: 1}
    assert execState.getTasksForExecutor(1) == {"t2"}
    assert execState.removeExecutor(1)
    assert execState._internals()


def test_execQueues():
    """test of ExecutorQueues"""
    for y in range(2):
        for i in range(3):
            assert eQ.pushTask("type%s" % y, "t%s%s" % (y, i)) == i + 1
    assert "DONE IN"
    res_internals = eQ._internals()
    assert res_internals["queues"] == {"type0": ["t00", "t01", "t02"], "type1": ["t10", "t11", "t12"]}
    assert set(res_internals["lastUse"].keys()) == {"type0", "type1"}
    assert res_internals["taskInQueue"] == {
        "t00": "type0",
        "t01": "type0",
        "t02": "type0",
        "t10": "type1",
        "t11": "type1",
        "t12": "type1",
    }
    assert eQ.pushTask("type0", "t01") == 3
    assert eQ.getState()
    assert eQ.popTask("type0")[0] == "t00"
    assert eQ.pushTask("type0", "t00", ahead=True) == 3
    assert eQ.popTask("type0")[0] == "t00"
    assert eQ.deleteTask("t01")
    res_internals = eQ._internals()
    assert res_internals["queues"] == {"type0": ["t02"], "type1": ["t10", "t11", "t12"]}
    assert set(res_internals["lastUse"].keys()) == {"type0", "type1"}
    assert res_internals["taskInQueue"] == {
        "t02": "type0",
        "t10": "type1",
        "t11": "type1",
        "t12": "type1",
    }
    assert eQ.getState()
    assert eQ.deleteTask("t02")
    res_internals = eQ._internals()
    assert res_internals["queues"] == {"type0": [], "type1": ["t10", "t11", "t12"]}
    assert set(res_internals["lastUse"].keys()) == {"type0", "type1"}
    assert res_internals["taskInQueue"] == {
        "t10": "type1",
        "t11": "type1",
        "t12": "type1",
    }
    assert eQ.getState()
    for i in range(3):
        assert eQ.popTask("type1")[0] == "t1%s" % i
    res_internals = eQ._internals()
    assert res_internals["queues"] == {"type0": [], "type1": []}
    assert set(res_internals["lastUse"].keys()) == {"type0", "type1"}
    assert res_internals["taskInQueue"] == {}

    assert eQ.pushTask("type0", "t00") == 1
    assert eQ.popTask("type0") == ("t00", "type0")
    res_internals = eQ._internals()
    assert res_internals["queues"] == {"type0": [], "type1": []}
    assert set(res_internals["lastUse"].keys()) == {"type0", "type1"}
    assert res_internals["taskInQueue"] == {}

    assert eQ.pushTask("type0", "t00") == 1
    assert eQ.deleteTask("t00")
    res_internals = eQ._internals()
    assert res_internals["queues"] == {"type0": [], "type1": []}
    assert set(res_internals["lastUse"].keys()) == {"type0", "type1"}
    assert res_internals["taskInQueue"] == {}

    assert not eQ.deleteTask("t00")



########################################################################
# File :   BOINCComputingElement.py
# Author : J.Wu
########################################################################

""" BOINC Computing Element
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import os
import bz2
import base64
import tempfile

from six.moves.urllib.parse import urlparse

from DIRAC import S_OK, S_ERROR
from DIRAC.Resources.Computing.ComputingElement import ComputingElement

CE_NAME = "BOINC"


class BOINCComputingElement(ComputingElement):

    ###############################################################################
    def __init__(self, ceUniqueID):
        """Standard constructor."""

        super(BOINCComputingElement, self).__init__(ceUniqueID)

        self.ceType = CE_NAME
        self.mandatoryParameters = []
        self.wsdl = None
        self.BOINCClient = None
        # define a job prefix based on the wsdl url
        self.suffix = None

    # this is for standlone test
    #    self.ceParameters['projectURL'] = 'http://mardirac3.in2p3.fr:7788/?wsdl'
    #    self.ceParameters['Platform'] = 'Linux_x86_64_glibc-2.5'

    def createClient(self):
        """
        This method only can be called after the initialisation of this class. In this
        method, it will initial some variables and create a soap client for communication
        with BOINC server.
        """

        if not self.wsdl:
            self.wsdl = self.ceParameters["projectURL"]
        if not self.suffix:
            result = urlparse(self.wsdl)
            self.suffix = result.hostname
        if not self.BOINCClient:
            try:
                from suds.client import Client

                if self.log.getLevel() == "DEBUG":
                    self.log.enableLogsFromExternalLibs()
                self.BOINCClient = Client(self.wsdl)
            except Exception as x:
                self.log.error("Creation of the soap client failed", "%s" % str(x))
                pass

    def submitJob(self, executableFile, proxy=None, numberOfJobs=1):
        """Method to submit job"""
        self.createClient()
        # Check if the client is ready
        if not self.BOINCClient:
            return S_ERROR("Soap client is not ready")

        self.log.verbose("Executable file path: %s" % executableFile)

        # if no proxy is supplied, the executable can be submitted directly
        # otherwise a wrapper script is needed to get the proxy to the execution node
        # The wrapper script makes debugging more complicated and thus it is
        # recommended to transfer a proxy inside the executable if possible.
        wrapperContent = ""
        if proxy:
            self.log.verbose("Setting up proxy for payload")

            compressedAndEncodedProxy = base64.b64encode(bz2.compress(proxy.dumpAllToString()["Value"])).decode()
            with open(executableFile, "rb") as fp:
                compressedAndEncodedExecutable = base64.b64encode(bz2.compress(fp.read(), 9)).decode()

            wrapperContent = """#!/bin/bash
/usr/bin/env python << EOF
# Wrapper script for executable and proxy
import os
import tempfile
import sys
import base64
import bz2
import shutil
import stat
try:
  workingDirectory = tempfile.mkdtemp( suffix = '_wrapper', prefix= 'TORQUE_' )
  os.chdir( workingDirectory )
  open( 'proxy', "w" ).write(bz2.decompress( base64.b64decode( "%(compressedAndEncodedProxy)s" ) ) )
  open( '%(executable)s', "w" ).write(bz2.decompress( base64.b64decode( "%(compressedAndEncodedExecutable)s" ) ) )
  os.chmod('proxy',stat.S_IRUSR | stat.S_IWUSR)
  os.chmod('%(executable)s',stat.S_IRWXU)
  os.environ["X509_USER_PROXY"]=os.path.join(workingDirectory, 'proxy')
except Exception as x:
  print >> sys.stderr, x
  sys.exit(-1)
cmd = "./%(executable)s"
print 'Executing: ', cmd
sys.stdout.flush()
os.system( cmd )

shutil.rmtree( workingDirectory )

EOF
""" % {
                "compressedAndEncodedProxy": compressedAndEncodedProxy,
                "compressedAndEncodedExecutable": compressedAndEncodedExecutable,
                "executable": os.path.basename(executableFile),
            }

            fd, name = tempfile.mkstemp(suffix="_pilotwrapper.py", prefix="DIRAC_", dir=os.getcwd())
            os.close(fd)
            submitFile = name

        else:  # no proxy
            submitFile = executableFile
            wrapperContent = self._fromFileToStr(submitFile)

        if not wrapperContent:
            self.log.error("Executable file is empty.")
            return S_ERROR("Executable file is empty.")

        # Some special symbol can not be transported by xml,
        # such as less, greater, amp. So, base64 is used here.
        wrapperContent = base64.b64encode(wrapperContent).decode()

        prefix = os.path.splitext(os.path.basename(submitFile))[0].replace("_pilotwrapper", "").replace("DIRAC_", "")
        batchIDList = []
        stampDict = {}
        for i in range(0, numberOfJobs):
            jobID = "%s_%d@%s" % (prefix, i, self.suffix)
            try:
                #  print jobID + "\n" + wrapperContent
                #  print self.BOINCClient
                result = self.BOINCClient.service.submitJob(
                    jobID, wrapperContent, self.ceParameters["Platform"][0], self.ceParameters["MarketPlaceID"]
                )
            except Exception:
                self.log.error(
                    "Could not submit the pilot to the BOINC CE", "Pilot %s, BOINC CE %s" % (jobID, self.wsdl)
                )
                break

            if not result["ok"]:
                self.log.warn(
                    "Didn't submit the pilot %s to the BOINC CE %s, the value returned is false!" % (jobID, self.wsdl)
                )
                break

            self.log.verbose("Submit the pilot %s to the BOINC CE %s" % (jobID, self.wsdl))
            diracStamp = "%s_%d" % (prefix, i)
            batchIDList.append(jobID)
            stampDict[jobID] = diracStamp

        if batchIDList:
            resultRe = S_OK(batchIDList)
            resultRe["PilotStampDict"] = stampDict
        else:
            resultRe = S_ERROR("Submit no pilot to BOINC CE %s" % self.wsdl)
        return resultRe

    #############################################################################
    def getCEStatus(self):
        """Method to get the BOINC CE dynamic jobs information."""
        self.createClient()
        # Check if the client is ready
        if not self.BOINCClient:
            self.log.error("Soap client is not ready.")
            return S_ERROR("Soap client is not ready.")

        try:
            result = self.BOINCClient.service.getDynamicInfo()
        except Exception:
            self.log.error("Could not get the BOINC CE dynamic jobs information", self.wsdl)
            return S_ERROR("Could not get the BOINC CE %s dynamic jobs information, communication failed!" % self.wsdl)

        if not result["ok"]:
            self.log.warn(
                "Did not get the BOINC CE %s dynamic jobs information, the value returned is false!" % self.wsdl
            )
            return S_ERROR(
                "Did not get the BOINC CE %s dynamic jobs information, the value returned is false!" % self.wsdl
            )

        self.log.verbose("Get the BOINC CE %s dynamic jobs info." % self.wsdl)

        resultRe = S_OK()
        resultRe["WaitingJobs"] = result["values"][0][0]
        resultRe["RunningJobs"] = result["values"][0][1]
        resultRe["SubmittedJobs"] = 0
        self.log.verbose("Waiting Jobs: ", resultRe["WaitingJobs"])
        self.log.verbose("Running Jobs: ", resultRe["RunningJobs"])
        return resultRe

    #############################################################################
    def getJobStatus(self, jobIDList):
        """Get the status information about jobs in the given list"""
        self.createClient()
        # Check if the client is ready
        if not self.BOINCClient:
            self.log.error("Soap client is not ready.")
            return S_ERROR("Soap client is not ready.")

        wsdl_jobIDList = self.BOINCClient.factory.create("stringArray")
        for job in jobIDList:
            try:
                job = job.split("@")[0]
            except Exception:
                self.log.debug("The job id is %s" % job)
                pass
            wsdl_jobIDList[0].append(job)

        try:
            result = self.BOINCClient.service.getJobStatus(wsdl_jobIDList)
        except Exception:
            self.log.error("Could not get the status about jobs in the list from the BOINC CE", self.wsdl)
            return S_ERROR(
                "Could not get the status about jobs in the list from the BOINC CE %s, commnication failed!" % self.wsdl
            )

        if not result["ok"]:
            self.log.warn(
                "Did not get the status about jobs in the list from the BOINC CE %s, the value returned is false!"
                % self.wsdl
            )
            return S_ERROR(
                "Did not get the status about jobs in the list from the BOINC CE %s, the value returned is false!"
                % self.wsdl
            )
        self.log.debug("Got the status about jobs in list from the BOINC CE %s." % self.wsdl)
        resultRe = {}
        for jobStatus in result["values"][0]:
            (jobID, status) = jobStatus.split(":")
            jobID = "%s@%s" % (jobID, self.suffix)
            resultRe[jobID] = status

        return S_OK(resultRe)

    #############################################################################
    def getJobOutput(self, jobID, localDir=None):
        """Get the stdout and stderr outputs of the specified job . If the localDir is provided,
        the outputs are stored as files in this directory and the name of the files are returned.
        Otherwise, the outputs are returned as strings.
        """
        self.createClient()
        # Check if the client is ready
        if not self.BOINCClient:
            self.log.error("Soap client is not ready.")
            return S_ERROR("Soap client is not ready.")

        try:
            tempID = jobID.split("@")[0]
        except Exception:
            tempID = jobID

        try:
            result = self.BOINCClient.service.getJobOutput(tempID)
        except Exception:
            self.log.error(
                "Could not get the outputs of job from the BOINC CE", "Job %s, BOINC CE %s" % (jobID, self.wsdl)
            )
            return S_ERROR(
                "Could not get the outputs of job %s from the BOINC CE %s, communication failed!" % (jobID, self.wsdl)
            )
        if not result["ok"]:
            self.log.warn(
                "Did not get the outputs of job %s from the BOINC CE %s, the value returned is false!"
                % (jobID, self.wsdl)
            )
            return S_ERROR(
                "Did not get the outputs of job %s from the BOINC CE %s, the value returned is false!"
                % (jobID, self.wsdl)
            )

        self.log.debug("Got the outputs of job %s from the BOINC CE %s." % (jobID, self.wsdl))

        strOutfile = base64.b64decode(result["values"][0][0])
        strErrorfile = base64.b64decode(result["values"][0][1])
        if localDir:
            outFile = os.path.join(localDir, "BOINC_%s.out" % jobID)
            self._fromStrToFile(strOutfile, outFile)

            errorFile = os.path.join(localDir, "BOINC_%s.err" % jobID)
            self._fromStrToFile(strErrorfile, errorFile)

            return S_OK((outFile, errorFile))
        else:
            # Return the outputs as a string
            return S_OK((strOutfile, strErrorfile))

    ##############################################################################
    def _fromFileToStr(self, fileName):
        """Read a file and return the file content as a string"""
        strFile = b""
        try:
            with open(fileName, "rb") as fileHander:
                strFile = fileHander.read()
        except Exception:
            self.log.verbose("To read file %s failed!\n" % fileName)
        return strFile

    #####################################################################
    def _fromStrToFile(self, strContent, fileName):
        """Write a string to a file"""
        try:
            with open(fileName, "wb") as fileHander:
                _ = fileHander.write(strContent)
        except Exception:
            self.log.verbose("To create %s failed!" % fileName)


# testing this
if __name__ == "__main__":

    test_boinc = BOINCComputingElement(12)
    test_submit = 1
    test_getStatus = 2
    test_getDynamic = 4
    test_getOutput = 8
    test_parameter = 4
    jobID = "zShvbK_0@mardirac3.in2p3.fr"
    if test_parameter & test_submit:
        fd, fname = tempfile.mkstemp(suffix="_pilotwrapper.py", prefix="DIRAC_", dir="/home/client/dirac/data/")
        os.close(fd)
        fd = open(fname, "w")
        fd.write('#!/usr/bin/env sh\necho "I am stadard out" &gt;&amp;1 \necho "I am stadard error" &gt;&amp;2 ')
        fd.close()
        result = test_boinc.submitJob(fname)

        if not result["OK"]:
            print(result["Message"])
        else:
            jobID = result["Value"][0]
            print("Successfully submit a job %s" % jobID)

    if test_parameter & test_getStatus:
        jobTestList = ["Uu0ghO_0@mardirac3.in2p3.fr", "1aDmIf_0@mardirac3.in2p3.fr", jobID]
        jobStatus = test_boinc.getJobStatus(jobTestList)
        if not jobStatus["OK"]:
            print(jobStatus["Message"])
        else:
            for _ in jobTestList:
                print("The status of the job %s is %s" % (id, jobStatus["Value"][id]))

    if test_parameter & test_getDynamic:
        serverState = test_boinc.getCEStatus()

        if not serverState["OK"]:
            print(serverState["Message"])
        else:
            print("The number of jobs waiting is %s" % serverState["WaitingJobs"])
            print("The number of jobs running is %s" % serverState["RunningJobs"])

    if test_parameter & test_getOutput:
        outstate = test_boinc.getJobOutput(jobID, "/tmp/")

        if not outstate["OK"]:
            print(outstate["Message"])
        else:
            print("Please check the directory /tmp for the output and error files of job %s" % jobID)



"""Unit tests of MQConsumer interface in the DIRAC.Resources.MessageQueue.MQConsumer
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
from DIRAC import S_OK
from DIRAC.Resources.MessageQueue.MQConsumer import MQConsumer
from DIRAC.Resources.MessageQueue.MQConnectionManager import MQConnectionManager
from DIRAC.Resources.MessageQueue.MQConnector import MQConnector


class FakeMQConnector(MQConnector):
    def __init__(self, params={}):
        super(FakeMQConnector, self).__init__()

    def disconnect(self):
        return S_OK("FakeMQConnection disconnecting")

    def get(self, destination=""):
        return "FakeMQConnection getting message"

    def subscribe(self, parameters=None):
        return S_OK("Subscription successful")

    def unsubscribe(self, parameters):
        return S_OK("Unsubscription successful")


class TestMQConsumer(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None  # To show full difference between structures in  case of error
        dest = {}
        dest.update({"/queue/FakeQueue": ["consumer4", "consumer2"]})
        dest4 = {"/queue/test3": ["consumer1", "consumer2", "consumer3", "consumer4"]}
        conn1 = {"MQConnector": FakeMQConnector(), "destinations": dest}
        conn2 = {"MQConnector": FakeMQConnector(), "destinations": dest4}
        storage = {"fake.cern.ch": conn1, "testdir.blabla.ch": conn2}
        self.myManager = MQConnectionManager(connectionStorage=storage)

    def tearDown(self):
        pass


class TestMQConsumer_get(TestMQConsumer):
    def test_failure(self):
        consumer = MQConsumer(mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", consumerId="consumer1")
        result = consumer.get()
        self.assertFalse(result["OK"])
        self.assertEqual(result["Message"], "No messages ( 1141 : No messages in queue)")

    def test_sucess(self):
        consumer = MQConsumer(mqManager=self.myManager, mqURI="bad.cern.ch::Queues::FakeQueue", consumerId="consumer1")
        result = consumer.get()
        self.assertFalse(result["OK"])


class TestMQConsumer_close(TestMQConsumer):
    def test_success(self):
        consumer = MQConsumer(mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", consumerId="consumer4")
        result = consumer.close()
        self.assertTrue(result["OK"])

    def test_failure(self):
        consumer = MQConsumer(mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", consumerId="consumer4")
        result = consumer.close()
        self.assertTrue(result["OK"])
        result = consumer.close()
        self.assertFalse(result["OK"])
        self.assertEqual(
            result["Message"],
            "MQ connection failure ( 1142 : Failed to stop the connection!The messenger consumer4 does not exist!)",
        )

    def test_failure2(self):
        consumer = MQConsumer(mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", consumerId="consumer4")
        consumer2 = MQConsumer(
            mqManager=self.myManager, mqURI="fake.cern.ch::Queues::FakeQueue", consumerId="consumer2"
        )
        result = consumer.close()
        self.assertTrue(result["OK"])
        result = consumer.close()
        self.assertFalse(result["OK"])
        self.assertEqual(
            result["Message"],
            "MQ connection failure ( 1142 : Failed to stop the connection!The messenger consumer4 does not exist!)",
        )
        result = consumer2.close()
        self.assertTrue(result["OK"])
        result = consumer2.close()
        self.assertFalse(result["OK"])
        self.assertEqual(
            result["Message"],
            "MQ connection failure ( 1142 : Failed to stop the connection!The messenger consumer2 does not exist!)",
        )


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConsumer)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConsumer_get))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(TestMQConsumer_close))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import mock
import unittest

import DIRAC
from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Resources.ProxyProvider.ProxyProviderFactory import ProxyProviderFactory


certsPath = os.path.join(os.path.dirname(DIRAC.__file__), "Core/Security/test/certs")


def sf_getInfoAboutProviders(of, providerName, option, section):
    if of == "Proxy" and option == "all" and section == "all":
        if providerName == "MY_DIRACCA":
            return S_OK(
                {
                    "ProviderType": "DIRACCA",
                    "CertFile": os.path.join(certsPath, "ca/ca.cert.pem"),
                    "KeyFile": os.path.join(certsPath, "ca/ca.key.pem"),
                    "Supplied": ["O", "OU", "CN"],
                    "Optional": ["emailAddress"],
                    "DNOrder": ["O", "OU", "CN", "emailAddress"],
                    "OU": "CA",
                    "C": "DN",
                    "O": "DIRACCA",
                }
            )
        elif providerName == "MY_PUSP":
            return S_OK({"ProviderType": "PUSP", "ServiceURL": "https://somedomain"})
    return S_ERROR("No proxy provider found")


@mock.patch("DIRAC.Resources.ProxyProvider.ProxyProviderFactory.getInfoAboutProviders", new=sf_getInfoAboutProviders)
class ProxyProviderFactoryTest(unittest.TestCase):
    """Base class for the ProxyProviderFactory test cases"""

    def test_standalone(self):
        """Test loading a proxy provider element with everything defined in itself."""
        for provider, resultOfGenerateDN in [("MY_DIRACCA", True), ("MY_PUSP", False)]:
            result = ProxyProviderFactory().getProxyProvider(provider)
            self.assertTrue(result["OK"], "\n" + result.get("Message", "Error message is absent."))
            proxyProviderObj = result["Value"]
            result = proxyProviderObj.generateDN(FullName="test", Email="email@test.org")
            text = "Must be ended %s%s" % (
                "successful" if resultOfGenerateDN else "with error",
                ": %s" % result.get("Message", "Error message is absent."),
            )
            self.assertEqual(result["OK"], resultOfGenerateDN, text)
            if not resultOfGenerateDN:
                gLogger.info("Msg: %s" % (result["Message"]))
            else:
                self.assertTrue(result["OK"], "\n" + result.get("Message", "Error message is absent."))
                userDN = result["Value"]
                gLogger.info("Created DN:", userDN)
                result = proxyProviderObj.getProxy(userDN)
                self.assertTrue(result["OK"], "\n" + result.get("Message", "Error message is absent."))


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ProxyProviderFactoryTest)
    testResult = unittest.TextTestRunner(verbosity=3).run(suite)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import pytest
from mock import MagicMock

from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getDIRACPlatform, getCompatiblePlatforms


mockGCReply = MagicMock()


@pytest.mark.parametrize(
    "mockGCReplyInput, requested, expectedRes, expectedValue",
    [
        ({"OK": False, "Message": "error"}, "plat", False, None),
        ({"OK": True, "Value": ""}, "plat", False, None),
        (
            {"OK": True, "Value": {"plat1": "OS1, OS2,  OS3", "plat2": "OS4, OS5", "plat3": "OS1, OS4"}},
            "plat",
            False,
            None,
        ),
        (
            {"OK": True, "Value": {"plat1": "OS1, OS2,  OS3", "plat2": "OS4, OS5", "plat3": "OS1, OS4"}},
            "OS1",
            True,
            ["plat1", "plat3"],
        ),
        (
            {"OK": True, "Value": {"plat1": "OS1, OS2,  OS3", "plat2": "OS4, OS5", "plat3": "OS1, OS4"}},
            "OS2",
            True,
            ["plat1"],
        ),
        (
            {"OK": True, "Value": {"plat1": "OS1, OS2,  OS3", "plat2": "OS4, OS5", "plat3": "OS1, OS4"}},
            "OS3",
            True,
            ["plat1"],
        ),
        (
            {"OK": True, "Value": {"plat1": "OS1, OS2,  OS3", "plat2": "OS4, OS5", "plat3": "OS1, OS4"}},
            "OS4",
            True,
            ["plat2", "plat3"],
        ),
        (
            {"OK": True, "Value": {"plat1": "OS1, OS2,  OS3", "plat2": "OS4, OS5", "plat3": "OS1, OS4"}},
            "OS5",
            True,
            ["plat2"],
        ),
        (
            {"OK": True, "Value": {"plat1": "OS1, OS2,  OS3", "plat2": "OS4, OS5", "plat3": "OS1, OS4"}},
            "plat1",
            True,
            ["plat1"],
        ),
    ],
)
def test_getDIRACPlatform(mocker, mockGCReplyInput, requested, expectedRes, expectedValue):

    mockGCReply.return_value = mockGCReplyInput

    mocker.patch("DIRAC.Interfaces.API.Dirac.gConfig.getOptionsDict", side_effect=mockGCReply)

    res = getDIRACPlatform(requested)
    assert res["OK"] is expectedRes, res
    if expectedRes:
        assert set(res["Value"]) == set(expectedValue), res["Value"]


@pytest.mark.parametrize(
    "mockGCReplyInput, requested, expectedRes, expectedValue",
    [
        ({"OK": False, "Message": "error"}, "plat", False, None),
        ({"OK": True, "Value": ""}, "plat", False, None),
        (
            {
                "OK": True,
                "Value": {"plat1": "xOS1, xOS2,  xOS3", "plat2": "sys2, xOS4, xOS5", "plat3": "sys1, xOS1, xOS4"},
            },
            "plat",
            True,
            ["plat"],
        ),
        (
            {
                "OK": True,
                "Value": {"plat1": "xOS1, xOS2,  xOS3", "plat2": "sys2, xOS4, xOS5", "plat3": "sys1, xOS1, xOS4"},
            },
            "plat1",
            True,
            ["plat1", "xOS1", "xOS2", "xOS3"],
        ),
    ],
)
def test_getCompatiblePlatforms(mocker, mockGCReplyInput, requested, expectedRes, expectedValue):
    mockGCReply.return_value = mockGCReplyInput

    mocker.patch("DIRAC.Interfaces.API.Dirac.gConfig.getOptionsDict", side_effect=mockGCReply)

    res = getCompatiblePlatforms(requested)
    assert res["OK"] is expectedRes, res
    if expectedRes:
        assert set(res["Value"]) == set(expectedValue), res["Value"]



""" Class that contains client access to the PilotsLogging handler. """
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Base.Client import Client, createClient


@createClient("WorkloadManagement/PilotsLogging")
class PilotsLoggingClient(Client):
    """Implementation of interface of Pilots Logging service. Client class should be used to communicate
    with PilotsLogging Service
    """

    def __init__(self, **kwargs):
        super(PilotsLoggingClient, self).__init__(**kwargs)
        self.setServer("WorkloadManagement/PilotsLogging")

    def addPilotsLogging(self, pilotUUID, timestamp, source, phase, status, messageContent):
        """
        Add new Pilots Logging entry

        :param pilotUUID: Pilot reference
        :param status: Pilot status
        :param minorStatus: Additional status information
        :param timestamp: Date and time of status event
        :param source: Source of status information
        """

        return self._getRPC().addPilotsLogging(pilotUUID, timestamp, source, phase, status, messageContent)

    def deletePilotsLogging(self, pilotUUID):
        """
        Delete all Logging entries for Pilot

        :param pilotUUID: Pilot reference
        """

        return self._getRPC().deletePilotsLogging(pilotUUID)

    def getPilotsLogging(self, pilotUUID):
        """
        Get all Logging entries for Pilot

        :param pilotUUID: Pilot reference
        """

        return self._getRPC().getPilotsLogging(pilotUUID)



""" pytest(s) for Executors
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

import pytest
from mock import MagicMock

from DIRAC.WorkloadManagementSystem.Client.JobState.CachedJobState import CachedJobState
from DIRAC.WorkloadManagementSystem.Client.JobState.JobManifest import JobManifest

# sut
from DIRAC.WorkloadManagementSystem.Executor.JobScheduling import JobScheduling
from DIRAC.WorkloadManagementSystem.Executor.InputData import InputData

mockNone = MagicMock()
mockNone.return_value = None


@pytest.mark.parametrize(
    "sites, banned, expected",
    [
        (["MY.Site1.org", "MY.Site2.org"], None, ["MY.Site1.org", "MY.Site2.org"]),
        (["MY.Site1.org", "MY.Site2.org"], ["MY.Site1.org", "MY.Site2.org"], []),
        (["MY.Site1.org", "MY.Site2.org"], ["MY.Site2.org"], ["MY.Site1.org"]),
        (["MY.Site1.org", "MY.Site2.org"], [], ["MY.Site1.org", "MY.Site2.org"]),
        ([], ["MY.Site1.org"], []),
        ([], [], []),
        (["MY.Site1.org", "MY.Site2.org"], ["MY.Site1.org"], ["MY.Site2.org"]),
        (["MY.Site1.org", "MY.Site2.org"], ["MY.Site1.org", "MY.Site3.org"], ["MY.Site2.org"]),
        ([], ["MY.Site1.org", "MY.Site3.org"], []),
        (["MY.Site1.org", "MY.Site2.org"], ["MY.Site4.org"], ["MY.Site1.org", "MY.Site2.org"]),
        (
            ["MY.Site1.org", "MY.Site2.org", "MY.Site3.org"],
            ["MY.Site4.org"],
            ["MY.Site1.org", "MY.Site2.org", "MY.Site3.org"],
        ),
        (["MY.Site1.org", "MY.Site2.org"], ["MY.Site4.org"], ["MY.Site1.org", "MY.Site2.org"]),
    ],
)
def test__applySiteFilter(sites, banned, expected):
    js = JobScheduling()
    filtered = js._applySiteFilter(sites, banned)
    assert set(filtered) == set(expected)


@pytest.mark.parametrize(
    "manifestOptions, expected",
    [
        ({}, []),
        ({"Tag": "bof"}, ["bof"]),
        ({"Tags": "bof, bif"}, ["bof", "bif"]),
        ({"MaxRAM": 2}, ["2GB"]),
        ({"Tags": "bof, bif", "MaxRAM": 2}, ["bof", "bif", "2GB"]),
        ({"WholeNode": "yes", "MaxRAM": 2}, ["WholeNode", "MultiProcessor", "2GB"]),
        ({"NumberOfProcessors": 1}, []),
        ({"NumberOfProcessors": 4}, ["MultiProcessor", "4Processors"]),
        ({"NumberOfProcessors": 4, "MinNumberOfProcessors": 2}, ["MultiProcessor", "4Processors"]),
        ({"NumberOfProcessors": 4, "MaxNumberOfProcessors": 12}, ["MultiProcessor", "4Processors"]),
        ({"NumberOfProcessors": 4, "MaxNumberOfProcessors": 12}, ["MultiProcessor", "4Processors"]),
        ({"MinNumberOfProcessors": 4, "MaxNumberOfProcessors": 12}, ["MultiProcessor", "4Processors"]),
        ({"MinNumberOfProcessors": 4, "MaxNumberOfProcessors": 4}, ["MultiProcessor", "4Processors"]),
        ({"MinNumberOfProcessors": 4}, ["MultiProcessor", "4Processors"]),
    ],
)
def test__getTagsFromManifest(manifestOptions, expected):
    manifest = JobManifest()
    for varName, varValue in manifestOptions.items():
        manifest.setOption(varName, varValue)

    js = JobScheduling()
    tagList = js._getTagsFromManifest(manifest)
    assert set(tagList) == set(expected)


@pytest.mark.parametrize(
    "manifestOptions, expected",
    [
        ({}, []),
        ({"InputSandbox": "SB:ProductionSandboxSE|/SandBox/l/lhcb_mc/4c7.bof"}, []),
        ({"InputSandbox": "LFN:/l/lhcb_mc/4c7.bof"}, ["/l/lhcb_mc/4c7.bof"]),
        (
            {"InputSandbox": ["SB:ProductionSandboxSE|/SandBox/l/lhcb_mc/4c7.bof", "LFN:/l/lhcb_mc/4c7.bof"]},
            ["/l/lhcb_mc/4c7.bof"],
        ),
        ({"InputSandbox": ["SB:ProductionSandboxSE|/SandBox/l/lhcb_mc/4c7.bof", "bif:/l/lhcb_mc/4c7.bof"]}, []),
        (
            {"InputSandbox": ["LFN:/l/lhcb_mc/4c8.bof", "LFN:/l/lhcb_mc/4c7.bof"]},
            ["/l/lhcb_mc/4c8.bof", "/l/lhcb_mc/4c7.bof"],
        ),
    ],
)
def test__getInputSandbox(mocker, manifestOptions, expected):

    mocker.patch("DIRAC.WorkloadManagementSystem.Client.JobState.JobState.JobDB.__init__", side_effect=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Client.JobState.JobState.JobLoggingDB.__init__", side_effect=mockNone)
    mocker.patch("DIRAC.WorkloadManagementSystem.Client.JobState.JobState.TaskQueueDB.__init__", side_effect=mockNone)

    inputData = InputData()
    js = CachedJobState(1)
    js.setAttribute("JobType", "User")

    manifest = JobManifest()
    for varName, varValue in manifestOptions.items():
        manifest.setOption(varName, varValue)

    js.setManifest(manifest)
    res = inputData._getInputSandbox(js)
    assert res["OK"] is True
    assert res["Value"] == expected



########################################################################
# File: Request.py
# Date: 2012/07/16 13:43:45
########################################################################
"""
:mod: Request

.. module: Request
  :synopsis: request implementation

request implementation
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Disable invalid names warning
# pylint: disable=invalid-name
__RCSID__ = "$Id$"


# # imports
import datetime
import json
import six

# # from DIRAC
from DIRAC import S_OK, S_ERROR
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.private.JSONUtils import RMSEncoder
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers


########################################################################
class Request(object):
    """
    :param int RequestID: requestID
    :param str Name: request' name
    :param str OwnerDN: request's owner DN
    :param str OwnerGroup: request owner group
    :param str Setup: DIRAC setup
    :param str SourceComponent: whatever
    :param int JobID: jobID
    :param datetime.datetime CreationTime: UTC datetime
    :param datetime.datetime SubmissionTime: UTC datetime
    :param datetime.datetime LastUpdate: UTC datetime
    :param datetime.datetime NotBefore: UTC datetime
    :param str Status: request's status
    :param TypedList.TypedList operations: list of operations

    It is managed by SQLAlchemy, so the RequestID should never be set by hand (except when constructed from
    JSON of course...)
    """

    ALL_STATES = ("Waiting", "Failed", "Done", "Scheduled", "Assigned", "Canceled")

    FINAL_STATES = ("Done", "Failed", "Canceled")

    _datetimeFormat = "%Y-%m-%d %H:%M:%S"

    def __init__(self, fromDict=None):
        """c'tor

        :param self: self reference
        :param fromDict: if false, new request. Can be json string that represents the object, or the dictionary directly
        """
        self.__waiting = None

        now = datetime.datetime.utcnow().replace(microsecond=0)

        self._CreationTime = now
        self._SubmitTime = now
        self._LastUpdate = now
        # the time before which the request should not be executed
        # If None, no delay
        self._NotBefore = now
        self._Status = "Done"
        self.JobID = 0
        self.Error = None
        self.DIRACSetup = None
        self.OwnerDN = None
        self.RequestName = None
        self.OwnerGroup = None
        self._SourceComponent = None

        self.dmsHelper = DMSHelpers()

        proxyInfo = getProxyInfo()
        if proxyInfo["OK"]:
            proxyInfo = proxyInfo["Value"]
            if proxyInfo["validGroup"] and proxyInfo["validDN"]:
                self.OwnerDN = proxyInfo["identity"]
                self.OwnerGroup = proxyInfo["group"]

        self.__operations__ = []

        if isinstance(fromDict, six.string_types):
            fromDict = json.loads(fromDict)
        elif not isinstance(fromDict, dict):
            fromDict = {}

        if "Operations" in fromDict:
            for opDict in fromDict.get("Operations", []):
                self += Operation(opDict)

            del fromDict["Operations"]

        for key, value in fromDict.items():
            # The JSON module forces the use of UTF-8, which is not properly
            # taken into account in DIRAC.
            # One would need to replace all the '== str' with 'in six.string_types'
            # This is converting `unicode` to `str` and doesn't make sense in Python 3
            if six.PY2 and isinstance(value, six.string_types):
                value = value.encode()

            if value:
                setattr(self, key, value)

        self._notify()

    def _notify(self):
        """simple state machine for sub request statuses"""
        # # update operations statuses
        self.__waiting = None

        # Update the Order in Operation, and set the parent
        for i in range(len(self.__operations__)):
            self.__operations__[i].Order = i
            self.__operations__[i]._parent = self

        rStatus = "Waiting"
        opStatusList = [(op.Status, op) for op in self]

        while opStatusList:
            # # Scan all status in order!
            opStatus, op = opStatusList.pop(0)

            # # Failed -> Failed
            if opStatus == "Failed" and self.__waiting is None:
                rStatus = "Failed"
                break

            # Scheduled -> Scheduled
            if opStatus == "Scheduled":
                if self.__waiting is None:
                    self.__waiting = op
                    rStatus = "Scheduled"
            # # First operation Queued becomes Waiting if no Waiting/Scheduled before
            elif opStatus == "Queued":
                if self.__waiting is None:
                    self.__waiting = op
                    op._setWaiting(self)
                    rStatus = "Waiting"
            # # First operation Waiting is next to execute, others are queued
            elif opStatus == "Waiting":
                rStatus = "Waiting"
                if self.__waiting is None:
                    self.__waiting = op
                else:
                    op._setQueued(self)
            # # All operations Done -> Done
            elif opStatus == "Done" and self.__waiting is None:
                rStatus = "Done"
                self.Error = ""
        self.Status = rStatus

    def getWaiting(self):
        """get waiting operation if any"""
        # # update states
        self._notify()
        return S_OK(self.__waiting)

    # # Operation arithmetics
    def __contains__(self, operation):
        """in operator

        :param self: self reference
        :param Operation.Operation subRequest: a subRequest
        """
        return bool(operation in self.__operations__)

    def __iadd__(self, operation):
        """+= operator for subRequest

        :param self: self reference
        :param Operation.Operation operation: sub-request to add
        """
        if operation not in self:
            self.__operations__.append(operation)
            operation._parent = self
            self._notify()
        return self

    def insertBefore(self, newOperation, existingOperation):
        """insert :newOperation: just before :existingOperation:

        :param self: self reference
        :param Operation.Operation newOperation: Operation to be inserted
        :param Operation.Operation existingOperation: previous Operation sibling
        """
        if existingOperation not in self:
            return S_ERROR("%s is not in" % existingOperation)
        if newOperation in self:
            return S_ERROR("%s is already in" % newOperation)
        self.__operations__.insert(self.__operations__.index(existingOperation), newOperation)
        self._notify()
        return S_OK()

    def insertAfter(self, newOperation, existingOperation):
        """insert :newOperation: just after :existingOperation:

        :param self: self reference
        :param Operation.Operation newOperation: Operation to be inserted
        :param Operation.Operation existingOperation: next Operation sibling
        """
        if existingOperation not in self:
            return S_ERROR("%s is not in" % existingOperation)
        if newOperation in self:
            return S_ERROR("%s is already in" % newOperation)
        self.__operations__.insert(self.__operations__.index(existingOperation) + 1, newOperation)
        self._notify()
        return S_OK()

    def addOperation(self, operation):
        """add :operation: to list of Operations

        :param self: self reference
        :param Operation.Operation operation: Operation to be inserted
        """
        if operation in self:
            return S_ERROR("This operation is already in!!!")
        self += operation
        return S_OK()

    def isEmpty(self):
        """Evaluate if the request is empty"""
        return len(self.__operations__) == 0

    def __iter__(self):
        """iterator for sub-request"""
        return self.__operations__.__iter__()

    def __getitem__(self, i):
        """[] op for sub requests"""
        return self.__operations__.__getitem__(i)

    def __setitem__(self, i, value):
        """self[i] = val"""

        self.__operations__.__setitem__(i, value)
        self._notify()

    def __delitem__(self, i):
        """del self[i]"""

        self.__operations__.__delitem__(i)
        self._notify()

    def indexOf(self, subReq):
        """return index of subReq (execution order)"""
        return self.__operations__.index(subReq) if subReq in self else -1

    def __bool__(self):
        """for comparisons"""
        return True

    # For Python 2 compatibility
    __nonzero__ = __bool__

    def __len__(self):
        """nb of subRequests"""
        return len(self.__operations__)

    def __str__(self):
        """str operator"""
        return self.toJSON()["Value"]

    def subStatusList(self):
        """list of statuses for all operations"""
        return [subReq.Status for subReq in self]

    @property
    def SourceComponent(self):
        return self._SourceComponent

    @SourceComponent.setter
    def SourceComponent(self, value):
        if isinstance(value, six.text_type):
            value = value.encode()
        if not isinstance(value, bytes):
            raise TypeError("SourceComponent should be bytes!")
        self._SourceComponent = value

    @property
    def CreationTime(self):
        """creation time getter"""
        return self._CreationTime

    @CreationTime.setter
    def CreationTime(self, value=None):
        """creation time setter"""
        if not isinstance(value, (datetime.datetime,) + six.string_types):
            raise TypeError("CreationTime should be a datetime.datetime!")
        if isinstance(value, six.string_types):
            value = datetime.datetime.strptime(value.split(".")[0], self._datetimeFormat)
        self._CreationTime = value

    @property
    def SubmitTime(self):
        """request's submission time getter"""
        return self._SubmitTime

    @SubmitTime.setter
    def SubmitTime(self, value=None):
        """submission time setter"""
        if not isinstance(value, (datetime.datetime,) + six.string_types):
            raise TypeError("SubmitTime should be a datetime.datetime!")
        if isinstance(value, six.string_types):
            value = datetime.datetime.strptime(value.split(".")[0], self._datetimeFormat)
        self._SubmitTime = value

    @property
    def NotBefore(self):
        """Getter for NotBefore time"""
        return self._NotBefore

    @NotBefore.setter
    def NotBefore(self, value=None):
        """Setter for the NotBefore time"""
        if not isinstance(value, (type(None), datetime.datetime) + six.string_types):
            raise TypeError("NotBefore should be a datetime.datetime!")
        if isinstance(value, six.string_types):
            value = datetime.datetime.strptime(value.split(".")[0], self._datetimeFormat)
        self._NotBefore = value

    def delayNextExecution(self, deltaTime):
        """This helper sets the NotBefore attribute in deltaTime minutes
        in the future

        :param deltaTime: time in minutes before next execution
        """
        now = datetime.datetime.utcnow().replace(microsecond=0)
        extraDelay = datetime.timedelta(minutes=deltaTime)
        self._NotBefore = now + extraDelay

        return S_OK()

    @property
    def LastUpdate(self):
        """last update getter"""
        return self._LastUpdate

    @LastUpdate.setter
    def LastUpdate(self, value=None):
        """last update setter"""
        if not isinstance(value, (datetime.datetime,) + six.string_types):
            raise TypeError("LastUpdate should be a datetime.datetime!")
        if isinstance(value, six.string_types):
            value = datetime.datetime.strptime(value.split(".")[0], self._datetimeFormat)
        self._LastUpdate = value

    @property
    def Status(self):
        """status getter"""
        self._notify()
        return self._Status

    @Status.setter
    def Status(self, value):
        """status setter"""
        if value not in Request.ALL_STATES:
            raise ValueError("Unknown status: %s" % str(value))

        # If the status moved to Failed or Done, update the lastUpdate time
        if value in ("Done", "Failed"):
            if value != self._Status:
                self.LastUpdate = datetime.datetime.utcnow().replace(microsecond=0)

        if value == "Done":
            self.Error = ""
        self._Status = value

    @property
    def Order(self):
        """ro execution order getter"""
        self._notify()
        opStatuses = [op.Status for op in self.__operations__]
        return opStatuses.index("Waiting") if "Waiting" in opStatuses else len(opStatuses)

    def toJSON(self):
        """Returns the JSON formated string that describes the request"""

        jsonStr = json.dumps(self, cls=RMSEncoder)
        return S_OK(jsonStr)

    def _getJSONData(self):
        """Returns the data that have to be serialized by JSON"""

        attrNames = [
            "RequestID",
            "RequestName",
            "OwnerDN",
            "OwnerGroup",
            "Status",
            "Error",
            "DIRACSetup",
            "SourceComponent",
            "JobID",
            "CreationTime",
            "SubmitTime",
            "LastUpdate",
            "NotBefore",
        ]
        jsonData = {}

        for attrName in attrNames:

            # RequestID might not be set since it is managed by SQLAlchemy
            if not hasattr(self, attrName):
                continue

            value = getattr(self, attrName)

            if isinstance(value, datetime.datetime):
                # We convert date time to a string
                jsonData[attrName] = value.strftime(self._datetimeFormat)  # pylint: disable=no-member
            else:
                jsonData[attrName] = value

        jsonData["Operations"] = self.__operations__

        return jsonData

    def getDigest(self):
        """return digest for request"""
        digest = ["Name:" + self.RequestName]
        for op in self:
            opDigest = [str(item) for item in (op.Type, op.Type, op.Status, op.Order)]
            if op.TargetSE:
                opDigest.append(op.TargetSE)
            if op.Catalog:
                opDigest.append(op.Catalog)
            if len(op):
                opFile = op[0]
                extraFilesStr = "...+<%d files>" % (len(op) - 1) if (len(op) > 1) else ""
                opDigest.append(opFile.LFN + extraFilesStr)
            digest.append(":".join(opDigest))
        return S_OK("\n".join(digest))

    def optimize(self):
        """Merges together the operations that can be merged. They need to have the following arguments equal:
        * Type
        * Arguments
        * SourceSE
        * TargetSE
        * Catalog
        It also makes sure that the maximum number of Files in an Operation is never overcome.

        CAUTION: this method is meant to be called before inserting into the DB.
                So if the RequestID is not 0, we don't touch

        :return: S_ERROR if the Request should not be optimized (because already in the DB)
                S_OK(True) if a optimization was carried out
                S_OK(False) if no optimization were carried out
        """

        # If the RequestID is not the default one (0), it probably means
        # the Request is already in the DB, so we don't touch anything
        if hasattr(self, "RequestID") and getattr(self, "RequestID"):
            return S_ERROR(
                "Cannot optimize because Request seems to be already in the DB (RequestID %s)"
                % getattr(self, "RequestID")
            )
        # Set to True if the request could be optimized
        optimized = False
        # Recognise Failover request series
        repAndRegList = []
        removeRepList = []
        i = 0
        while i < len(self.__operations__):
            insertNow = True
            if i < len(self.__operations__) - 1:
                op1 = self.__operations__[i]
                op2 = self.__operations__[i + 1]
                if getattr(op1, "Type") == "ReplicateAndRegister" and getattr(op2, "Type") == "RemoveReplica":
                    fileSetA = set(list(f.LFN for f in op1))
                    fileSetB = set(list(f.LFN for f in op2))
                    if fileSetA == fileSetB:
                        # Source is useless if failover
                        if self.dmsHelper.isSEFailover(op1.SourceSE):
                            op1.SourceSE = ""
                        repAndRegList.append((op1.TargetSE, op1))
                        removeRepList.append((op2.TargetSE, op2))
                        del self.__operations__[i]
                        del self.__operations__[i]
                        # If we are at the end of the request, we must insert the new operations
                        insertNow = i == len(self.__operations__)
            # print i, self.__operations__[i].Type if i < len( self.__operations__ )
            # else None, len( repAndRegList ), insertNow
            if insertNow:
                if repAndRegList:
                    # We must insert the new operations there
                    # i.e. insert before operation i (if it exists)
                    # Replication first, removeReplica next
                    optimized = True
                    insertBefore = self.__operations__[i] if i < len(self.__operations__) else None
                    # print 'Insert new operations before', insertBefore
                    ops = [op for _, op in sorted(repAndRegList, key=lambda x: x[0])]
                    ops += [op for _, op in sorted(removeRepList, key=lambda x: x[0])]
                    for op in ops:
                        _res = self.insertBefore(op, insertBefore) if insertBefore else self.addOperation(op)
                        # Skip the newly inserted operation
                        i += 1
                    repAndRegList = []
                    removeRepList = []
                else:
                    # Skip current operation
                    i += 1
            else:
                # Just to show that in that case we don't increment i
                pass

        # List of attributes that must be equal for operations to be merged
        attrList = ["Type", "Arguments", "SourceSE", "TargetSE", "Catalog"]

        i = 0
        while i < len(self.__operations__):
            while i < len(self.__operations__) - 1:
                # Some attributes need to be the same
                attrMismatch = False
                for attr in attrList:
                    if getattr(self.__operations__[i], attr) != getattr(self.__operations__[i + 1], attr):
                        attrMismatch = True
                        break

                if attrMismatch:
                    break

                # We do not do the merge if there are common files in the operations
                fileSetA = set(list(f.LFN for f in self.__operations__[i]))
                fileSetB = set(list(f.LFN for f in self.__operations__[i + 1]))
                if fileSetA & fileSetB:
                    break

                # There is a maximum number of files one can add into an operation
                try:
                    while len(self.__operations__[i + 1]):
                        fileToMove = self.__operations__[i + 1][0]
                        self.__operations__[i] += fileToMove

                        # If the object is mapped to SQLAlchemy object with a relationship
                        # having the delete-orphan option, the fileToMove will have
                        # already disappeared from the original operation. Silly...
                        # If not, we have to remove it manually

                        if len(self.__operations__[i + 1]) and (self.__operations__[i + 1][0] == fileToMove):
                            del self.__operations__[i + 1][0]
                        optimized = True
                    del self.__operations__[i + 1]
                except RuntimeError:
                    i += 1
            i += 1

        return S_OK(optimized)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import S_OK
from DIRAC.AccountingSystem.Client.Types.Pilot import Pilot
from DIRAC.AccountingSystem.private.Plotters.BaseReporter import BaseReporter


class PilotPlotter(BaseReporter):

    _typeName = "Pilot"
    _typeKeyFields = [dF[0] for dF in Pilot().definitionKeyFields]

    def _reportCumulativeNumberOfJobs(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "Jobs"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        dataDict = self._accumulate(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, maxValue, unitName = self._findSuitableUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "jobs"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCumulativeNumberOfJobs(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Cumulative Jobs by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
            "sort_labels": "last_value",
        }
        return self._generateCumulativePlot(filename, plotInfo["graphDataDict"], metadata)

    def _reportNumberOfJobs(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "Jobs"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "jobs"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotNumberOfJobs(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Jobs by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"],
        }
        return self._generateTimedStackedBarPlot(filename, plotInfo["graphDataDict"], metadata)

    def _reportCumulativeNumberOfPilots(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        dataDict = self._accumulate(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, maxValue, unitName = self._findSuitableUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "jobs"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotCumulativeNumberOfPilots(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Cumulative Pilots by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"].replace("job", "pilot"),
            "sort_labels": "last_value",
        }
        return self._generateCumulativePlot(filename, plotInfo["graphDataDict"], metadata)

    def _reportNumberOfPilots(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict, maxValue = self._divideByFactor(dataDict, granularity)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        baseDataDict, graphDataDict, maxValue, unitName = self._findSuitableRateUnit(
            dataDict, self._getAccumulationMaxValue(dataDict), "jobs"
        )
        return S_OK(
            {"data": baseDataDict, "graphDataDict": graphDataDict, "granularity": granularity, "unit": unitName}
        )

    def _plotNumberOfPilots(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Pilots by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": plotInfo["unit"].replace("job", "pilot"),
        }
        return self._generateTimedStackedBarPlot(filename, plotInfo["graphDataDict"], metadata)

    def _reportJobsPerPilot(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", %s, %s, SUM(%s), SUM(%s)",
            reportRequest["groupingFields"][1] + ["startTime", "bucketLength", "Jobs", "entriesInBucket"],
        )
        retVal = self._getTimedData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {
                "checkNone": True,
                "convertToGranularity": "sum",
                "calculateProportionalGauges": False,
                "consolidationFunction": self._averageConsolidation,
            },
        )
        if not retVal["OK"]:
            return retVal
        dataDict, granularity = retVal["Value"]
        self.stripDataField(dataDict, 0)
        dataDict = self._fillWithZero(granularity, reportRequest["startTime"], reportRequest["endTime"], dataDict)
        return S_OK({"data": dataDict, "granularity": granularity})

    def _plotJobsPerPilot(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Jobs per pilot by %s" % reportRequest["grouping"],
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
            "span": plotInfo["granularity"],
            "ylabel": "jobs/pilot",
            "normalization": max(x for y in plotInfo["data"].values() for x in y.values()),
        }
        return self._generateQualityPlot(filename, plotInfo["data"], metadata)

    def _reportTotalNumberOfPilots(self, reportRequest):
        selectFields = (
            self._getSelectStringForGrouping(reportRequest["groupingFields"]) + ", SUM(%s)",
            reportRequest["groupingFields"][1] + ["entriesInBucket"],
        )
        retVal = self._getSummaryData(
            reportRequest["startTime"],
            reportRequest["endTime"],
            selectFields,
            reportRequest["condDict"],
            reportRequest["groupingFields"],
            {},
        )
        if not retVal["OK"]:
            return retVal
        dataDict = retVal["Value"]
        return S_OK({"data": dataDict})

    def _plotTotalNumberOfPilots(self, reportRequest, plotInfo, filename):
        metadata = {
            "title": "Total Number of Pilots by %s" % reportRequest["grouping"],
            "ylabel": "Pilots",
            "starttime": reportRequest["startTime"],
            "endtime": reportRequest["endTime"],
        }
        return self._generatePiePlot(filename, plotInfo["data"], metadata)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from DIRAC.AccountingSystem.Client.Types.BaseAccountingType import BaseAccountingType
import DIRAC

__RCSID__ = "$Id$"


class DataOperation(BaseAccountingType):
    def __init__(self):
        super(DataOperation, self).__init__()
        self.definitionKeyFields = [
            ("OperationType", "VARCHAR(32)"),
            ("User", "VARCHAR(64)"),
            ("ExecutionSite", "VARCHAR(256)"),
            ("Source", "VARCHAR(32)"),
            ("Destination", "VARCHAR(32)"),
            ("Protocol", "VARCHAR(32)"),
            ("FinalStatus", "VARCHAR(32)"),
        ]
        self.definitionAccountingFields = [
            ("TransferSize", "BIGINT UNSIGNED"),
            ("TransferTime", "FLOAT"),
            ("RegistrationTime", "FLOAT"),
            ("TransferOK", "INT UNSIGNED"),
            ("TransferTotal", "INT UNSIGNED"),
            ("RegistrationOK", "INT UNSIGNED"),
            ("RegistrationTotal", "INT UNSIGNED"),
        ]
        self.bucketsLength = [
            (86400 * 3, 900),  # <3d = 15m
            (86400 * 8, 3600),  # <1w+1d = 1h
            (15552000, 86400),  # >1w+1d <6m = 1d
            (31104000, 604800),  # >6m = 1w
        ]
        self.checkType()
        self.setValueByKey("ExecutionSite", DIRAC.siteName())



"""
Unit tests of utility functions in the DIRAC.Resources.MessageQueue.Utilities
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import DIRAC.Resources.MessageQueue.Utilities as module
import unittest
from six.moves import queue as Queue

from mock import MagicMock

__RCSID__ = "$Id$"


ROOT_PATH = "/Resources/MQServices/"
MQSERVICE_NAME = "mq.dirac.net"
QUEUE_TYPE = "Queue"
TOPIC_TYPE = "Topic"
QUEUE_NAME = "Test"

CS_MQSERVICE_OPTIONS = {"Host": MQSERVICE_NAME}
CS_QUEUE_OPTIONS = {"Acknowledgement": True}

QUEUE_CONFIG = {
    "%s/%s/Queues/%s" % (ROOT_PATH, MQSERVICE_NAME, QUEUE_NAME): None,
    "%s/%s/Queues/%s/Acknowledgement" % (ROOT_PATH, MQSERVICE_NAME, QUEUE_NAME): True,
}

TOPIC_CONFIG = {
    "%s/%s/Topics/%s" % (ROOT_PATH, MQSERVICE_NAME, QUEUE_NAME): None,
    "%s/%s/Topics/%s/Acknowledgement" % (ROOT_PATH, MQSERVICE_NAME, QUEUE_NAME): True,
}

SIMILAR_QUEUE_CONFIG = QUEUE_CONFIG.copy()
SIMILAR_QUEUE_CONFIG.update(
    {
        "%s/%s/Queues/%s1" % (ROOT_PATH, MQSERVICE_NAME, QUEUE_NAME): None,
        "%s/%s/Queues/%s1/Acknowledgement" % (ROOT_PATH, MQSERVICE_NAME, QUEUE_NAME): True,
    }
)

DIFFERENT_MQSERVICE_NAME = "different-mq.dirac.net"
AMBIGIOUS_QUEUE_CONFIG = QUEUE_CONFIG.copy()
AMBIGIOUS_QUEUE_CONFIG.update(
    {
        "%s/%s/Queues/%s" % (ROOT_PATH, DIFFERENT_MQSERVICE_NAME, QUEUE_NAME): None,
        "%s/%s/Queues/%s/Acknowledgement" % (ROOT_PATH, DIFFERENT_MQSERVICE_NAME, QUEUE_NAME): True,
    }
)

"""
Used CS example:
Resources
{
  MQServices
  {
    mq.dirac.net
    {
      Host = mq.dirac.net
      Queues
      {
        Test
        {
          Acknowledgement = True
        }
      }
      Topics
      {
        Test
        {
          Acknowledgement = True
        }
      }
    }
    different-mq.dirac.net
    {
      Host = different-mq.dirac.net
      Queues
      {
        Test
        {
          Acknowledgement = True
        }
      }
      Topics
      {
        Test
        {
          Acknowledgement = True
        }
      }
    }
  }
}
"""


class Test_getMQParamFromCSSuccessTestCase(unittest.TestCase):
    """Test class to check success scenarios."""

    def setUp(self):

        # external dependencies
        module.CSAPI = MagicMock()
        module.gConfig = MagicMock()

        module.gConfig.getOptionsDict.side_effect = [
            {"OK": True, "Value": CS_MQSERVICE_OPTIONS},
            {"OK": True, "Value": CS_QUEUE_OPTIONS},
        ]

    def test_getQueue(self):

        module.gConfig.getConfigurationTree.return_value = {"OK": True, "Value": QUEUE_CONFIG}

        # check returned value
        result = module.getMQParamsFromCS(mqURI=MQSERVICE_NAME + "::" + QUEUE_TYPE + "::" + QUEUE_NAME)
        self.assertTrue(result["OK"])

        # check queue parameters
        self.assertEqual(result["Value"]["Queue"], QUEUE_NAME)
        self.assertEqual(result["Value"]["Host"], MQSERVICE_NAME)
        self.assertTrue(result["Value"]["Acknowledgement"])

        with self.assertRaises(KeyError):
            result["Value"]["Topic"]

    def test_getTopic(self):
        """Try to get a topic"""
        module.gConfig.getConfigurationTree.return_value = {"OK": True, "Value": TOPIC_CONFIG}

        # check returned value
        TOPIC_NAME = QUEUE_NAME
        result = module.getMQParamsFromCS(mqURI=MQSERVICE_NAME + "::" + TOPIC_TYPE + "::" + TOPIC_NAME)
        self.assertTrue(result["OK"])

        # check topic parameters
        self.assertEqual(result["Value"]["Topic"], TOPIC_NAME)
        self.assertEqual(result["Value"]["Host"], MQSERVICE_NAME)
        self.assertTrue(result["Value"]["Acknowledgement"])

        with self.assertRaises(KeyError):
            result["Value"]["Queue"]

    def test_getMQService(self):
        self.assertEqual(module.getMQService("bblabl.ch::Topics::MyTopic"), "bblabl.ch")
        self.assertEqual(module.getMQService("bblabl.ch::Queues::MyQueue"), "bblabl.ch")

    def test_getDestinationType(self):
        self.assertEqual(module.getDestinationType("bblabl.ch::Topics::MyTopic"), "Topics")
        self.assertEqual(module.getDestinationType("bblabl.ch::Queues::MyQueue"), "Queues")

    def test_getDestinationName(self):
        self.assertEqual(module.getDestinationName("bblabl.ch::Topics::MyTopic"), "MyTopic")
        self.assertEqual(module.getDestinationName("bblabl.ch::Queues::MyQueue"), "MyQueue")

    def test_getDestinationAddress(self):
        self.assertEqual(module.getDestinationAddress("bblabl.ch::Topics::MyTopic"), "/topic/MyTopic")
        self.assertEqual(module.getDestinationAddress("bblabl.ch::Queues::MyQueue"), "/queue/MyQueue")


class Test_getMQParamFromCSFailureTestCase(unittest.TestCase):
    """Test class to check known failure scenarios."""

    def setUp(self):

        # external dependencies
        module.CSAPI = MagicMock()
        module.gConfig = MagicMock()

    def test_getQueueByInvalidName(self):
        """Try to get a queue by invalid names"""

        module.gConfig.getConfigurationTree.return_value = {"OK": True, "Value": {}}

        # try different possibilities
        result = module.getMQParamsFromCS("%s" % QUEUE_NAME)
        self.assertFalse(result["OK"])

        result = module.getMQParamsFromCS(mqURI=MQSERVICE_NAME + "::" + QUEUE_TYPE + "::" + "InvalidName")
        self.assertFalse(result["OK"])

        result = module.getMQParamsFromCS("%s::" % MQSERVICE_NAME)
        self.assertFalse(result["OK"])

        result = module.getMQParamsFromCS("%s::%s" % (MQSERVICE_NAME, QUEUE_NAME))
        self.assertFalse(result["OK"])


class Test_generateDefaultCallbackTestCase(unittest.TestCase):
    """Check default callback behaviour."""

    def test_EmptyMessage(self):
        myCallback = module.generateDefaultCallback()
        self.assertRaises(Queue.Empty, myCallback.get)

    def test_putOneGetOneMessage(self):
        myCallback = module.generateDefaultCallback()
        myCallback("", "test message")
        self.assertEqual(myCallback.get(), "test message")

    def test_severalMessages(self):
        myCallback = module.generateDefaultCallback()
        myCallback("", "test message1")
        myCallback("", "test message2")
        myCallback("", "test message3")
        myCallback("", "test message4")
        self.assertEqual(myCallback.get(), "test message1")
        self.assertEqual(myCallback.get(), "test message2")
        self.assertEqual(myCallback.get(), "test message3")
        self.assertEqual(myCallback.get(), "test message4")
        self.assertRaises(Queue.Empty, myCallback.get)

    def test_twoDifferentCallbacks(self):
        myCallback = module.generateDefaultCallback()
        myCallback2 = module.generateDefaultCallback()
        myCallback("", "test message")
        myCallback2("", "test message2")
        self.assertEqual(myCallback.get(), "test message")
        self.assertRaises(Queue.Empty, myCallback.get)
        self.assertEqual(myCallback2.get(), "test message2")
        self.assertRaises(Queue.Empty, myCallback2.get)


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(Test_getMQParamFromCSSuccessTestCase)
    suite.addTests(unittest.defaultTestLoader.loadTestsFromTestCase(Test_getMQParamFromCSFailureTestCase))
    suite.addTests(unittest.defaultTestLoader.loadTestsFromTestCase(Test_generateDefaultCallbackTestCase))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
########################################################################
# File :    dirac-admin-get-banned-sites
# Author :  Stuart Paterson
########################################################################
"""
Get banned sites

Example:
  $ dirac-admin-get-banned-sites
  LCG.IN2P3.fr                      Site not present in logging table
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    Script.parseCommandLine(ignoreErrors=True)

    from DIRAC import gLogger, exit as DIRACExit
    from DIRAC.Interfaces.API.DiracAdmin import DiracAdmin

    diracAdmin = DiracAdmin()

    result = diracAdmin.getBannedSites()
    if result["OK"]:
        bannedSites = result["Value"]
    else:
        gLogger.error(result["Message"])
        DIRACExit(2)

    for site in bannedSites:
        result = diracAdmin.getSiteMaskLogging(site, printOutput=True)
        if not result["OK"]:
            gLogger.error(result["Message"])
            DIRACExit(2)

    DIRACExit(0)


if __name__ == "__main__":
    main()



"""Test the dirac-transformation-replication script and helper"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from mock import MagicMock as Mock, patch

from DIRAC import S_OK, S_ERROR

from DIRAC.TransformationSystem.Utilities.ReplicationTransformation import createDataTransformation
from DIRAC.TransformationSystem.Utilities.ReplicationCLIParameters import Params

__RCSID__ = "$Id$"

GET_VOMS = "DIRAC.TransformationSystem.Utilities.ReplicationCLIParameters.getVOMSVOForGroup"
GET_PROXY = "DIRAC.TransformationSystem.Utilities.ReplicationCLIParameters.getProxyInfo"


def getProxyMock(success=True):
    """return value for getProxy"""
    if success:
        return Mock(
            return_value=S_OK(
                {
                    "groupProperties": ["ProductionManagement"],
                    "group": "clic_prod",
                }
            )
        )
    return Mock(return_value=S_ERROR("Failed"))


def opMock():
    """return mock for config operations"""
    opmock = Mock()
    opmock.getOptionsDict.return_value = S_OK({"trans": "ProdID"})
    opmock.getValue.return_value = "ProdID"
    return Mock(return_value=opmock)


class TestMoving(unittest.TestCase):
    """Test the creation of moving transformation"""

    def setUp(self):
        self.tClientMock = Mock()
        # self.tClientMock.createTransformationInputDataQuery.return_value = S_OK()
        self.tClientMock.createTransformationMetaQuery.return_value = S_OK()
        self.tMock = Mock(return_value=self.tClientMock)

    def tearDown(self):
        pass

    def test_createRepl_Broadcast(self):
        """test creating transformation"""
        tSE = "Target-SRM"
        sSE = "Source-SRM"
        prodID = 12345
        trmodule = "DIRAC.TransformationSystem.Client.Transformation.Transformation"
        with patch(trmodule + ".getTransformation", new=Mock(return_value=S_OK({}))), patch(
            trmodule + ".addTransformation", new=Mock(return_value=S_OK())
        ), patch(trmodule + "._Transformation__setSE", new=Mock(return_value=S_OK())):
            ret = createDataTransformation("Moving", tSE, sSE, "prodID", prodID, enable=True)
        self.assertTrue(ret["OK"], ret.get("Message", ""))
        self.assertEqual(ret["Value"].getPlugin().get("Value"), "Broadcast")
        self.assertEqual(ret["Value"].inputMetaQuery, {"prodID": prodID})

    def test_createRepl_NoSource(self):
        """test creating transformation"""
        tSE = "Target-SRM"
        sSE = ""
        prodID = 12345
        trmodule = "DIRAC.TransformationSystem.Client.Transformation.Transformation"
        with patch(trmodule + ".getTransformation", new=Mock(return_value=S_OK({}))), patch(
            trmodule + ".addTransformation", new=Mock(return_value=S_OK())
        ), patch(trmodule + "._Transformation__setSE", new=Mock(return_value=S_OK())):
            ret = createDataTransformation("Moving", tSE, sSE, "prodID", prodID, enable=True)
        self.assertTrue(ret["OK"], ret.get("Message", ""))
        self.assertEqual(ret["Value"].getPlugin().get("Value"), "Broadcast")

    def test_createRepl_Dry(self):
        """test creating transformation"""
        tSE = "Target-SRM"
        sSE = "Source-SRM"
        prodID = 12345
        trmodule = "DIRAC.TransformationSystem.Client.Transformation.Transformation"
        with patch(trmodule + ".getTransformation", new=Mock(return_value=S_OK({}))), patch(
            trmodule + ".addTransformation", new=Mock(return_value=S_OK())
        ), patch(trmodule + "._Transformation__setSE", new=Mock(return_value=S_OK())):
            ret = createDataTransformation("Moving", tSE, sSE, "prodID", prodID, enable=False, extraData={})
        self.assertTrue(ret["OK"], ret.get("Message", ""))

    def test_createRepl_2(self):
        """test creating transformation"""
        tSE = "Target-SRM"
        sSE = "Source-SRM"
        prodID = 12345
        trmodule = "DIRAC.TransformationSystem.Client.Transformation.Transformation"
        with patch(trmodule + ".getTransformation", new=Mock(return_value=S_OK({}))), patch(
            trmodule + ".addTransformation", new=Mock(return_value=S_OK())
        ), patch(trmodule + "._Transformation__setSE", new=Mock(return_value=S_OK())):
            ret = createDataTransformation("Moving", tSE, sSE, "prodID", prodID, extraname="extraName", enable=True)
        self.assertTrue(ret["OK"], ret.get("Message", ""))

    def test_createRepl_BadFlavour(self):
        """test creating transformation"""
        tSE = "Target-SRM"
        sSE = "Source-SRM"
        prodID = 12345
        trmodule = "DIRAC.TransformationSystem.Client.Transformation.Transformation"
        with patch(trmodule + ".getTransformation", new=Mock(return_value=S_OK({}))), patch(
            trmodule + ".addTransformation", new=Mock(return_value=S_OK())
        ), patch(trmodule + "._Transformation__setSE", new=Mock(return_value=S_OK())):
            ret = createDataTransformation("Smelly", tSE, sSE, "prodID", prodID, extraname="extraName", enable=True)
        self.assertFalse(ret["OK"], ret.get("Message", ""))
        self.assertIn("Unsupported flavour", ret["Message"])

    def test_createRepl_SEFail_1(self):
        """test creating transformation"""
        tSE = "Target-SRM"
        sSE = "Source-SRM"
        prodID = 12345
        trmodule = "DIRAC.TransformationSystem.Client.Transformation.Transformation"
        with patch(trmodule + ".getTransformation", new=Mock(return_value=S_OK({}))), patch(
            trmodule + ".addTransformation", new=Mock(return_value=S_OK())
        ), patch(trmodule + "._Transformation__setSE", new=Mock(side_effect=(S_OK(), S_ERROR()))):
            ret = createDataTransformation("Moving", tSE, sSE, "prodID", prodID, enable=True)
        self.assertFalse(ret["OK"], str(ret))
        self.assertIn("TargetSE not valid", ret["Message"])

    def test_createRepl_SEFail_2(self):
        """test creating transformation"""
        tSE = "Target-SRM"
        sSE = "Source-SRM"
        prodID = 12345
        trmodule = "DIRAC.TransformationSystem.Client.Transformation.Transformation"
        with patch(trmodule + ".getTransformation", new=Mock(return_value=S_OK({}))), patch(
            trmodule + ".addTransformation", new=Mock(return_value=S_OK())
        ), patch(trmodule + "._Transformation__setSE", new=Mock(side_effect=(S_ERROR(), S_ERROR()))):
            ret = createDataTransformation("Moving", tSE, sSE, "prodID", prodID, enable=True)
        self.assertFalse(ret["OK"], str(ret))
        self.assertIn("SourceSE not valid", ret["Message"])

    def test_createRepl_addTrafoFail_(self):
        """test creating transformation"""
        tSE = "Target-SRM"
        sSE = "Source-SRM"
        prodID = 12345
        trmodule = "DIRAC.TransformationSystem.Client.Transformation.Transformation"
        with patch(trmodule + ".getTransformation", new=Mock(return_value=S_OK({}))), patch(
            trmodule + ".addTransformation", new=Mock(return_value=S_ERROR("Cannot add Trafo"))
        ), patch(trmodule + "._Transformation__setSE", new=Mock(return_value=S_OK())):
            ret = createDataTransformation("Moving", tSE, sSE, "prodID", prodID, enable=True)
        self.assertFalse(ret["OK"], str(ret))
        self.assertIn("Cannot add Trafo", ret["Message"])

    def test_createRepl_createTrafoFail_(self):
        """test creating transformation"""
        tSE = "Target-SRM"
        sSE = "Source-SRM"
        prodID = 12345
        trmodule = "DIRAC.TransformationSystem.Client.Transformation.Transformation"
        with patch(trmodule + ".getTransformation", new=Mock(return_value=S_OK({}))), patch(
            trmodule + ".addTransformation", new=Mock(return_value=S_ERROR("Failed to add IMQ"))
        ), patch(trmodule + "._Transformation__setSE", new=Mock(return_value=S_OK())):
            ret = createDataTransformation("Moving", tSE, sSE, "prodID", prodID, enable=True)
        self.assertFalse(ret["OK"], str(ret))
        self.assertIn("Failed to add IMQ", ret["Message"])


class TestParams(unittest.TestCase):
    """Test the parameters for the moving creation script"""

    def setUp(self):
        self.arguments = []
        self.sMock = Mock()
        self.sMock.getPositionalArgs.return_value = self.arguments
        self.params = Params()

    def tearDown(self):
        pass

    @patch(GET_PROXY, new=getProxyMock())
    @patch(GET_VOMS, new=Mock(return_value="clic"))
    def test_checkSettings(self):
        self.arguments = ["12345", "TargetSE"]
        self.sMock.getPositionalArgs.return_value = self.arguments
        ret = self.params.checkSettings(self.sMock)
        self.assertTrue(ret["OK"], ret.get("Message", ""))
        self.assertEqual(self.params.metaValues, ["12345"])
        self.assertEqual(self.params.sourceSE, "")
        self.assertEqual(self.params.targetSE, ["TargetSE"])

    @patch(GET_PROXY, new=getProxyMock())
    @patch(GET_VOMS, new=Mock(return_value="clic"))
    def test_setMetadata(self):
        ret = self.params.setMetadata("Datatype:GEN, Energy: 124")
        self.assertTrue(ret["OK"], ret.get("Message", ""))
        self.assertEqual(self.params.extraData, {"Datatype": "GEN", "Energy": "124"})

    @patch(GET_PROXY, new=getProxyMock())
    @patch(GET_VOMS, new=Mock(return_value="clic"))
    def test_checkSettings_FailArgumentSize(self):
        self.arguments = ["12345", "TargetSE", "Foo"]
        self.sMock.getPositionalArgs.return_value = self.arguments
        ret = self.params.checkSettings(self.sMock)
        self.assertFalse(ret["OK"], str(ret))
        self.assertTrue(any("ERROR: Wrong number of arguments" in msg for msg in self.params.errorMessages))

    @patch(GET_PROXY, new=getProxyMock(False))
    @patch(GET_VOMS, new=Mock(return_value="clic"))
    def test_FailProxy(self):
        self.arguments = ["12345", "TargetSE"]
        self.sMock.getPositionalArgs.return_value = self.arguments
        ret = self.params.checkSettings(self.sMock)
        self.assertFalse(ret["OK"], str(ret))
        self.assertTrue(
            any("ERROR: No Proxy" in msg for msg in self.params.errorMessages), str(self.params.errorMessages)
        )

    @patch(GET_PROXY, new=getProxyMock(True))
    @patch(GET_VOMS, new=Mock(return_value=""))
    def test_FailProxy2(self):
        self.arguments = ["12345", "TargetSE"]
        self.sMock.getPositionalArgs.return_value = self.arguments
        ret = self.params.checkSettings(self.sMock)
        self.assertFalse(ret["OK"], str(ret))
        self.assertTrue(
            any("ERROR: ProxyGroup" in msg for msg in self.params.errorMessages), str(self.params.errorMessages)
        )

    def test_setExtraName(self):
        ret = self.params.setExtraname("extraName")
        self.assertTrue(ret["OK"], ret.get("Message", ""))
        self.assertEqual("extraName", self.params.extraname)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger
from DIRAC.Core.Tornado.Server.TornadoService import TornadoService
from DIRAC.DataManagementSystem.Service.FTS3ManagerHandler import FTS3ManagerHandlerMixin

sLog = gLogger.getSubLogger(__name__)


class TornadoFTS3ManagerHandler(FTS3ManagerHandlerMixin, TornadoService):
    """Tornado handler for the FTS3Manager"""

    log = sLog



""" DIRAC Workload Management System utility module to get available memory and processors from mjf
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import re
import multiprocessing
from six.moves.urllib.request import urlopen

from DIRAC import gLogger, gConfig
from DIRAC.Core.Utilities.List import fromChar


def getJobFeatures():
    features = {}
    if "JOBFEATURES" not in os.environ:
        return features
    for item in (
        "allocated_cpu",
        "hs06_job",
        "shutdowntime_job",
        "grace_secs_job",
        "jobstart_secs",
        "job_id",
        "wall_limit_secs",
        "cpu_limit_secs",
        "max_rss_bytes",
        "max_swap_bytes",
        "scratch_limit_bytes",
    ):
        fname = os.path.join(os.environ["JOBFEATURES"], item)
        try:
            val = urlopen(fname).read()
        except Exception:
            val = 0
        features[item] = val
    return features


def getProcessorFromMJF():
    jobFeatures = getJobFeatures()
    if jobFeatures:
        try:
            return int(jobFeatures["allocated_cpu"])
        except KeyError:
            gLogger.error(
                "MJF is available but allocated_cpu is not an integer", repr(jobFeatures.get("allocated_cpu"))
            )
    return None


def getMemoryFromMJF():
    jobFeatures = getJobFeatures()
    if jobFeatures:
        try:
            return int(jobFeatures["max_rss_bytes"])
        except KeyError:
            gLogger.error(
                "MJF is available but max_rss_bytes is not an integer", repr(jobFeatures.get("max_rss_bytes"))
            )
    return None


def getMemoryFromProc():
    meminfo = dict((i.split()[0].rstrip(":"), int(i.split()[1])) for i in open("/proc/meminfo").readlines())
    maxRAM = meminfo["MemTotal"]
    if maxRAM:
        return int(maxRAM / 1024)


def getNumberOfProcessors(siteName=None, gridCE=None, queue=None):
    """gets the number of processors on a certain CE/queue/node (what the pilot administers)

    The siteName/gridCE/queue parameters are normally not necessary.

    Tries to find it in this order:
    1) from the /Resources/Computing/CEDefaults/NumberOfProcessors (which is what the pilot fills up)
    2) if not present from JobFeatures
    3) if not present looks in CS for "NumberOfProcessors" Queue or CE option
    4) if not present but there's WholeNode tag, look what the WN provides using multiprocessing.cpu_count()
    5) return 1
    """

    # 1) from /Resources/Computing/CEDefaults/NumberOfProcessors
    gLogger.info("Getting numberOfProcessors from /Resources/Computing/CEDefaults/NumberOfProcessors")
    numberOfProcessors = gConfig.getValue("/Resources/Computing/CEDefaults/NumberOfProcessors", 0)
    if numberOfProcessors:
        return numberOfProcessors

    # 2) from MJF
    gLogger.info("Getting numberOfProcessors from MJF")
    numberOfProcessors = getProcessorFromMJF()
    if numberOfProcessors:
        return numberOfProcessors
    gLogger.info("NumberOfProcessors could not be found in MJF")

    # 3) looks in CS for "NumberOfProcessors" Queue or CE or site option
    if not siteName:
        siteName = gConfig.getValue("/LocalSite/Site", "")
    if not gridCE:
        gridCE = gConfig.getValue("/LocalSite/GridCE", "")
    if not queue:
        queue = gConfig.getValue("/LocalSite/CEQueue", "")
    if not (siteName and gridCE and queue):
        gLogger.error("Could not find NumberOfProcessors: missing siteName or gridCE or queue. Returning '1'")
        return 1

    grid = siteName.split(".")[0]
    csPaths = [
        "/Resources/Sites/%s/%s/CEs/%s/Queues/%s/NumberOfProcessors" % (grid, siteName, gridCE, queue),
        "/Resources/Sites/%s/%s/CEs/%s/NumberOfProcessors" % (grid, siteName, gridCE),
        "/Resources/Sites/%s/%s/Cloud/%s/VMTypes/%s/NumberOfProcessors" % (grid, siteName, gridCE, queue),
        "/Resources/Sites/%s/%s/Cloud/%s/NumberOfProcessors" % (grid, siteName, gridCE),
        "/Resources/Sites/%s/%s/NumberOfProcessors" % (grid, siteName),
    ]
    for csPath in csPaths:
        gLogger.info("Looking in", csPath)
        numberOfProcessors = gConfig.getValue(csPath, 0)
        if numberOfProcessors:
            return numberOfProcessors

    # 4) looks in CS for tags
    gLogger.info("Getting tags" "for %s: %s: %s" % (siteName, gridCE, queue))
    # Tags of the CE
    tags = fromChar(
        gConfig.getValue("/Resources/Sites/%s/%s/CEs/%s/Tag" % (siteName.split(".")[0], siteName, gridCE), "")
    ) + fromChar(
        gConfig.getValue("/Resources/Sites/%s/%s/Cloud/%s/Tag" % (siteName.split(".")[0], siteName, gridCE), "")
    )
    # Tags of the Queue
    tags += fromChar(
        gConfig.getValue(
            "/Resources/Sites/%s/%s/CEs/%s/Queues/%s/Tag" % (siteName.split(".")[0], siteName, gridCE, queue), ""
        )
    ) + fromChar(
        gConfig.getValue(
            "/Resources/Sites/%s/%s/Cloud/%s/VMTypes/%s/Tag" % (siteName.split(".")[0], siteName, gridCE, queue), ""
        )
    )
    gLogger.info("NumberOfProcessors could not be found in CS")
    if "WholeNode" in tags:
        gLogger.info("Found WholeNode tag, using multiprocessing.cpu_count()")
        return multiprocessing.cpu_count()

    # 5) return the default
    return 1


def getNumberOfPayloadProcessors(siteName=None, gridCE=None, queue=None):
    """Gets the number of processors allowed for a single JobAgent (so for a "inner" CE).
    (NB: this does not refer to the job processors).
    This is normally used ONLY when a pilot instantiates more than one JobAgent (MultiLaunchAgent pilot command).

    The siteName/gridCE/queue parameters are normally not necessary.

    Tries to find it in this order:
    1) from the /Resources/Computing/CEDefaults/NumberOfPayloadProcessors (which is what pilot 3 fills up)
    2) if not present but there's WholeNode tag, use the getNumberOfProcessors function above
    3) otherwise returns 1
    """

    # 1) from /Resources/Computing/CEDefaults/NumberOfPayloadProcessors
    gLogger.info("Getting NumberOfPayloadProcessors from /Resources/Computing/CEDefaults/NumberOfPayloadProcessors")
    NumberOfPayloadProcessors = gConfig.getValue("/Resources/Computing/CEDefaults/NumberOfPayloadProcessors")
    if NumberOfPayloadProcessors:
        return NumberOfPayloadProcessors

    # 2) Checks if 'WholeNode' is one of the used tags
    # Tags of the CE
    tags = fromChar(
        gConfig.getValue("/Resources/Sites/%s/%s/CEs/%s/Tag" % (siteName.split(".")[0], siteName, gridCE), "")
    ) + fromChar(
        gConfig.getValue("/Resources/Sites/%s/%s/Cloud/%s/Tag" % (siteName.split(".")[0], siteName, gridCE), "")
    )
    # Tags of the Queue
    tags += fromChar(
        gConfig.getValue(
            "/Resources/Sites/%s/%s/CEs/%s/Queues/%s/Tag" % (siteName.split(".")[0], siteName, gridCE, queue), ""
        )
    ) + fromChar(
        gConfig.getValue(
            "/Resources/Sites/%s/%s/Cloud/%s/VMTypes/%s/Tag" % (siteName.split(".")[0], siteName, gridCE, queue), ""
        )
    )

    if "WholeNode" in tags:
        return getNumberOfProcessors()

    # 3) Just returns a conservative "1"
    return 1


def getNumberOfJobProcessors(jobID):
    """Gets the number of processors allowed for the job.
    This can be used to communicate to your job payload the number of processors it's allowed to use,
    so this function should be called from your extension.

    If the JobAgent is using "InProcess" CE (which is the default),
    then what's returned will basically be the same of what's returned by the getNumberOfProcessors() function above
    """

    # from /Resources/Computing/JobLimits/jobID/NumberOfProcessors (set by PoolComputingElement)
    numberOfProcessors = gConfig.getValue("Resources/Computing/JobLimits/%s/NumberOfProcessors" % jobID)
    if numberOfProcessors:
        return numberOfProcessors

    return getNumberOfProcessors()


def getNumberOfGPUs(siteName=None, gridCE=None, queue=None):
    """Gets GPUs on a certain CE/queue/node (what the pilot administers)

    The siteName/gridCE/queue parameters are normally not necessary.

    Tries to find it in this order:
    1) from the /Resources/Computing/CEDefaults/GPUs (which is what the pilot might fill up)
    2) if not present looks in CS for "NumberOfGPUs" Queue or CE option
    3) return 0
    """

    # 1) from /Resources/Computing/CEDefaults/NumberOfGPUs
    gLogger.info("Getting GPUs from /Resources/Computing/CEDefaults/NumberOfGPUs")
    gpus = gConfig.getValue("/Resources/Computing/CEDefaults/NumberOfGPUs", 0)
    if gpus:
        return gpus

    # 2) looks in CS for "NumberOfGPUs" Queue or CE or site option
    if not siteName:
        siteName = gConfig.getValue("/LocalSite/Site", "")
    if not gridCE:
        gridCE = gConfig.getValue("/LocalSite/GridCE", "")
    if not queue:
        queue = gConfig.getValue("/LocalSite/CEQueue", "")
    if not (siteName and gridCE and queue):
        gLogger.error("Could not find NumberOfGPUs: missing siteName or gridCE or queue. Returning '0'")
        return 0

    grid = siteName.split(".")[0]
    csPaths = [
        "/Resources/Sites/%s/%s/CEs/%s/Queues/%s/NumberOfGPUs" % (grid, siteName, gridCE, queue),
        "/Resources/Sites/%s/%s/CEs/%s/NumberOfGPUs" % (grid, siteName, gridCE),
        "/Resources/Sites/%s/%s/Cloud/%s/VMTypes/%s/NumberOfGPUs" % (grid, siteName, gridCE, queue),
        "/Resources/Sites/%s/%s/Cloud/%s/NumberOfGPUs" % (grid, siteName, gridCE),
        "/Resources/Sites/%s/%s/NumberOfGPUs" % (grid, siteName),
    ]
    for csPath in csPaths:
        gLogger.info("Looking in", csPath)
        numberOfGPUs = gConfig.getValue(csPath, 0)
        if numberOfGPUs:
            return numberOfGPUs

    # 3) return 0
    gLogger.info("NumberOfGPUs could not be found in CS")
    return 0



""" MQConsumer
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from six.moves import queue as Queue
from DIRAC import S_ERROR, S_OK, gLogger
from DIRAC.Resources.MessageQueue.Utilities import getDestinationAddress, getMQService, generateDefaultCallback
from DIRAC.Core.Utilities.DErrno import EMQNOM


class MQConsumer(object):
    def __init__(self, mqManager, mqURI, consumerId, callback=generateDefaultCallback()):
        self._connectionManager = mqManager
        self._mqURI = mqURI
        self._destination = getDestinationAddress(self._mqURI)
        self._id = consumerId
        self._callback = callback
        self.log = gLogger.getSubLogger(self.__class__.__name__)
        # subscribing to connection
        result = self._connectionManager.getConnector(getMQService(self._mqURI))
        if result["OK"]:
            connector = result["Value"]
            if connector:
                result = connector.subscribe(
                    parameters={"messengerId": self._id, "callback": callback, "destination": self._destination}
                )
                if not result["OK"]:
                    self.log.error("Failed to subscribe the consumer:" + self._id)
            else:
                self.log.error("Failed to initialize MQConsumer! No MQConnector!")
        else:
            self.log.error("Failed to get MQConnector!")

    def get(self):
        """Function gets the message
            using the default callback machinery.
            This function can be called only if the the default
            callback function was used !!!!

        Returns:
          S_OK or S_ERROR: Error in case if there are no messages in the
            queue or other error appeared.
            S_OK with the message content otherwise.
        """
        if not self._callback:
            return S_ERROR("No callback set!")
        try:
            msg = self._callback.get()
        except Queue.Empty:
            return S_ERROR(EMQNOM, "No messages in queue")
        except Exception as e:
            return S_ERROR("Exception: %s" % e)
        else:
            return S_OK(msg)

    def close(self):
        """Function closes the connection for this client.
            The producer id is removed from the connection storage.
            It is not guaranteed that the connection will be
            removed cause other messengers can be still using it.

        Returns:
          S_OK or S_ERROR: Error appears in case if the connection was already
            closed for this consumer.
        """
        return self._connectionManager.stopConnection(mqURI=self._mqURI, messengerId=self._id)



""" This module exposes the BaseClient class,
    which serves as base for InnerRPCClient and TransferClient.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import six
import time

from six.moves import _thread as thread

import DIRAC
from DIRAC.Core.DISET.private.Protocols import gProtocolDict
from DIRAC.FrameworkSystem.Client.Logger import gLogger
from DIRAC.Core.Utilities import List, Network
from DIRAC.Core.Utilities.ReturnValues import S_OK, S_ERROR
from DIRAC.ConfigurationSystem.Client.Config import gConfig
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceURL, getServiceFailoverURL
from DIRAC.ConfigurationSystem.Client.Helpers import Registry
from DIRAC.ConfigurationSystem.Client.Helpers.CSGlobals import skipCACheck
from DIRAC.Core.DISET.private.TransportPool import getGlobalTransportPool
from DIRAC.Core.DISET.ThreadConfig import ThreadConfig


class BaseClient(object):
    """Glues together stubs with threading, credentials, and URLs discovery (by DIRAC vo and setup).
    Basically what needs to be done to enable RPC calls, and transfer, to find a URL.
    """

    VAL_EXTRA_CREDENTIALS_HOST = "hosts"

    KW_USE_CERTIFICATES = "useCertificates"
    KW_EXTRA_CREDENTIALS = "extraCredentials"
    KW_TIMEOUT = "timeout"
    KW_SETUP = "setup"
    KW_VO = "VO"
    KW_DELEGATED_DN = "delegatedDN"
    KW_DELEGATED_GROUP = "delegatedGroup"
    KW_IGNORE_GATEWAYS = "ignoreGateways"
    KW_PROXY_LOCATION = "proxyLocation"
    KW_PROXY_STRING = "proxyString"
    KW_PROXY_CHAIN = "proxyChain"
    KW_SKIP_CA_CHECK = "skipCACheck"
    KW_KEEP_ALIVE_LAPSE = "keepAliveLapse"

    __threadConfig = ThreadConfig()

    def __init__(self, serviceName, **kwargs):
        """Constructor

        :param serviceName: URL of the service (proper uri or just System/Component)
        :param useCertificates: If set to True, use the server certificate
        :param extraCredentials:
        :param timeout: Timeout of the call (default 600 s)
        :param setup: Specify the Setup
        :param VO: Specify the VO
        :param delegatedDN: Not clear what it can be used for.
        :param delegatedGroup: Not clear what it can be used for.
        :param ignoreGateways: Ignore the DIRAC Gatways settings
        :param proxyLocation: Specify the location of the proxy
        :param proxyString: Specify the proxy string
        :param proxyChain: Specify the proxy chain
        :param skipCACheck: Do not check the CA
        :param keepAliveLapse: Duration for keepAliveLapse (heartbeat like)
        """

        if not isinstance(serviceName, six.string_types):
            raise TypeError(
                "Service name expected to be a string. Received %s type %s" % (str(serviceName), type(serviceName))
            )
        # Explicitly convert to a str to avoid Python 2 M2Crypto issues with unicode objects
        self._destinationSrv = str(serviceName)
        self._serviceName = str(serviceName)
        self.kwargs = kwargs
        self.__useCertificates = None
        # The CS useServerCertificate option can be overridden by explicit argument
        self.__forceUseCertificates = self.kwargs.get(self.KW_USE_CERTIFICATES)
        self.__initStatus = S_OK()
        self.__idDict = {}
        self.__extraCredentials = ""
        self.__enableThreadCheck = False
        self.__retry = 0
        self.__retryDelay = 0
        # by default we always have 1 url for example:
        # RPCClient('dips://volhcb38.cern.ch:9162/Framework/SystemAdministrator')
        self.__nbOfUrls = 1
        self.__nbOfRetry = 3  # by default we try try times
        self.__retryCounter = 1
        self.__bannedUrls = []
        for initFunc in (
            self.__discoverSetup,
            self.__discoverVO,
            self.__discoverTimeout,
            self.__discoverURL,
            self.__discoverCredentialsToUse,
            self.__checkTransportSanity,
            self.__setKeepAliveLapse,
        ):
            result = initFunc()
            if not result["OK"] and self.__initStatus["OK"]:
                self.__initStatus = result
        self.numberOfURLs = 0
        self._initialize()
        # HACK for thread-safety:
        self.__allowedThreadID = False

    def _initialize(self):
        pass

    def getDestinationService(self):
        """Return service destination

        :return: str
        """
        return self._destinationSrv

    def getServiceName(self):
        """Return service name

        :return: str
        """
        return self._serviceName

    def __discoverSetup(self):
        """Discover which setup to use and stores it in self.setup
        The setup is looked for:
           * kwargs of the constructor (see KW_SETUP)
           * the ThreadConfig
           * in the CS /DIRAC/Setup
           * default to 'Test'

        :return: S_OK()/S_ERROR()
        """
        if self.KW_SETUP in self.kwargs and self.kwargs[self.KW_SETUP]:
            self.setup = str(self.kwargs[self.KW_SETUP])
        else:
            self.setup = self.__threadConfig.getSetup()
            if not self.setup:
                self.setup = gConfig.getValue("/DIRAC/Setup", "Test")
        return S_OK()

    def __discoverVO(self):
        """Discover which VO to use and stores it in self.vo
        The VO is looked for:
           * kwargs of the constructor (see KW_VO)
           * in the CS /DIRAC/VirtualOrganization
           * default to 'unknown'

        :return: S_OK()/S_ERROR()
        """
        if self.KW_VO in self.kwargs and self.kwargs[self.KW_VO]:
            self.vo = str(self.kwargs[self.KW_VO])
        else:
            self.vo = gConfig.getValue("/DIRAC/VirtualOrganization", "unknown")
        return S_OK()

    def __discoverURL(self):
        """Calculate the final URL. It is called at initialization and in connect in case of issue

        It sets:
          * self.serviceURL: the url (dips) selected as target using __findServiceURL
          * self.__URLTuple: a split of serviceURL obtained by Network.splitURL
          * self._serviceName: the last part of URLTuple (typically System/Component)

        :return: S_OK()/S_ERROR()
        """
        # Calculate final URL
        try:
            result = self.__findServiceURL()
        except Exception as e:
            return S_ERROR(repr(e))
        if not result["OK"]:
            return result
        self.serviceURL = result["Value"]
        retVal = Network.splitURL(self.serviceURL)
        if not retVal["OK"]:
            return retVal
        self.__URLTuple = retVal["Value"]
        self._serviceName = self.__URLTuple[-1]
        res = gConfig.getOptionsDict("/DIRAC/ConnConf/%s:%s" % self.__URLTuple[1:3])
        if res["OK"]:
            opts = res["Value"]
            for k in opts:
                if k not in self.kwargs:
                    self.kwargs[k] = opts[k]
        return S_OK()

    def __discoverTimeout(self):
        """Discover which timeout to use and stores it in self.timeout
        The timeout can be specified kwargs of the constructor (see KW_TIMEOUT),
        with a minimum of 120 seconds.
        If unspecified, the timeout will be 600 seconds.
        The value is set in self.timeout, as well as in self.kwargs[KW_TIMEOUT]

        :return: S_OK()/S_ERROR()
        """
        if self.KW_TIMEOUT in self.kwargs:
            self.timeout = self.kwargs[self.KW_TIMEOUT]
        else:
            self.timeout = False
        if self.timeout:
            self.timeout = max(120, self.timeout)
        else:
            self.timeout = 600
        self.kwargs[self.KW_TIMEOUT] = self.timeout
        return S_OK()

    def __discoverCredentialsToUse(self):
        """Discovers which credentials to use for connection.

        * Server certificate:
          -> If KW_USE_CERTIFICATES in kwargs, sets it in self.__useCertificates
          -> If not, check gConfig.useServerCertificate(),
             and sets it in self.__useCertificates and kwargs[KW_USE_CERTIFICATES]

        * Certification Authorities check:
           -> if KW_SKIP_CA_CHECK is not in kwargs and we are using the certificates,
              set KW_SKIP_CA_CHECK to false in kwargs
           -> if KW_SKIP_CA_CHECK is not in kwargs and we are not using the certificate, check the CS.skipCACheck

        * Proxy Chain
           -> if KW_PROXY_CHAIN in kwargs, we remove it and dump its string form into kwargs[KW_PROXY_STRING]

        :return: S_OK()/S_ERROR()
        """
        # Use certificates?
        if self.KW_USE_CERTIFICATES in self.kwargs:
            self.__useCertificates = self.kwargs[self.KW_USE_CERTIFICATES]
        else:
            self.__useCertificates = gConfig.useServerCertificate()
            self.kwargs[self.KW_USE_CERTIFICATES] = self.__useCertificates
        if self.KW_SKIP_CA_CHECK not in self.kwargs:
            if self.__useCertificates:
                self.kwargs[self.KW_SKIP_CA_CHECK] = False
            else:
                self.kwargs[self.KW_SKIP_CA_CHECK] = skipCACheck()
        if self.KW_PROXY_CHAIN in self.kwargs:
            try:
                self.kwargs[self.KW_PROXY_STRING] = self.kwargs[self.KW_PROXY_CHAIN].dumpAllToString()["Value"]
                del self.kwargs[self.KW_PROXY_CHAIN]
            except Exception:
                return S_ERROR("Invalid proxy chain specified on instantiation")
        return S_OK()

    def __discoverExtraCredentials(self):
        """Add extra credentials informations.
        * self.__extraCredentials
          -> if KW_EXTRA_CREDENTIALS in kwargs, we set it
          -> Otherwise, if we use the server certificate, we set it to VAL_EXTRA_CREDENTIALS_HOST
          -> If we have a delegation (see bellow), we set it to (delegatedDN, delegatedGroup)
          -> otherwise it is an empty string
        * delegation:
          -> if KW_DELEGATED_DN in kwargs, or delegatedDN in threadConfig, put in in self.kwargs
          -> if KW_DELEGATED_GROUP in kwargs or delegatedGroup in threadConfig, put it in self.kwargs
          -> If we have a delegated DN but not group, we find the corresponding group in the CS

        :return: S_OK()/S_ERROR()
        """
        # which extra credentials to use?
        self.__extraCredentials = self.VAL_EXTRA_CREDENTIALS_HOST if self.__useCertificates else ""
        if self.KW_EXTRA_CREDENTIALS in self.kwargs:
            self.__extraCredentials = self.kwargs[self.KW_EXTRA_CREDENTIALS]

        # Are we delegating something?
        delegatedDN = self.kwargs.get(self.KW_DELEGATED_DN) or self.__threadConfig.getDN()
        delegatedGroup = self.kwargs.get(self.KW_DELEGATED_GROUP) or self.__threadConfig.getGroup()
        if delegatedDN:
            self.kwargs[self.KW_DELEGATED_DN] = delegatedDN
            if not delegatedGroup:
                result = Registry.findDefaultGroupForDN(delegatedDN)
                if not result["OK"]:
                    return result
                delegatedGroup = result["Value"]
            self.kwargs[self.KW_DELEGATED_GROUP] = delegatedGroup
            self.__extraCredentials = (delegatedDN, delegatedGroup)
        return S_OK()

    def __findServiceURL(self):
        """Discovers the URL of a service, taking into account gateways, multiple URLs, banned URLs


        If the site on which we run is configured to use gateways (/DIRAC/Gateways/<siteName>),
        these URLs will be used. To ignore the gateway, it is possible to set KW_IGNORE_GATEWAYS
        to False in kwargs.

        If self._destinationSrv (given as constructor attribute) is a properly formed URL,
        we just return this one. If we have to use a gateway, we just replace the server name in the url.

        The list of URLs defined in the CS (<System>/URLs/<Component>) is randomized

        This method also sets some attributes:
          * self.__nbOfUrls = number of URLs
          * self.__nbOfRetry = 2 if we have more than 2 urls, otherwise 3
          * self.__bannedUrls is reinitialized if all the URLs are banned

        :return: S_OK(str)/S_ERROR() -- the selected URL
        """
        if not self.__initStatus["OK"]:
            return self.__initStatus

        # Load the Gateways URLs for the current site Name
        gatewayURL = False
        if not self.kwargs.get(self.KW_IGNORE_GATEWAYS):
            dRetVal = gConfig.getOption("/DIRAC/Gateways/%s" % DIRAC.siteName())
            if dRetVal["OK"]:
                rawGatewayURL = List.randomize(List.fromChar(dRetVal["Value"], ","))[0]
                gatewayURL = "/".join(rawGatewayURL.split("/")[:3])

        # If what was given as constructor attribute is a properly formed URL,
        # we just return this one.
        # If we have to use a gateway, we just replace the server name in it
        for protocol in gProtocolDict:
            if self._destinationSrv.find("%s://" % protocol) == 0:
                gLogger.debug("Already given a valid url", self._destinationSrv)
                if not gatewayURL:
                    return S_OK(self._destinationSrv)
                gLogger.debug("Reconstructing given URL to pass through gateway")
                path = "/".join(self._destinationSrv.split("/")[3:])
                finalURL = "%s/%s" % (gatewayURL, path)
                gLogger.debug("Gateway URL conversion:\n %s -> %s" % (self._destinationSrv, finalURL))
                return S_OK(finalURL)

        if gatewayURL:
            gLogger.debug("Using gateway", gatewayURL)
            return S_OK("%s/%s" % (gatewayURL, self._destinationSrv))

        # We extract the list of URLs from the CS (System/URLs/Component)
        try:
            urls = getServiceURL(self._destinationSrv, setup=self.setup)
        except Exception as e:
            return S_ERROR("Cannot get URL for %s in setup %s: %s" % (self._destinationSrv, self.setup, repr(e)))
        if not urls:
            return S_ERROR("URL for service %s not found" % self._destinationSrv)

        failoverUrls = []
        # Try if there are some failover URLs to use as last resort
        try:
            failoverUrlsStr = getServiceFailoverURL(self._destinationSrv, setup=self.setup)
            if failoverUrlsStr:
                failoverUrls = failoverUrlsStr.split(",")
        except Exception as e:
            pass

        # We randomize the list, and add at the end the failover URLs (System/FailoverURLs/Component)
        urlsList = List.randomize(List.fromChar(urls, ",")) + failoverUrls
        self.__nbOfUrls = len(urlsList)
        self.__nbOfRetry = (
            2 if self.__nbOfUrls > 2 else 3
        )  # we retry 2 times all services, if we run more than 2 services
        if self.__nbOfUrls == len(self.__bannedUrls):
            self.__bannedUrls = []  # retry all urls
            gLogger.debug("Retrying again all URLs")

        if len(self.__bannedUrls) > 0 and len(urlsList) > 1:
            # we have host which is not accessible. We remove that host from the list.
            # We only remove if we have more than one instance
            for i in self.__bannedUrls:
                gLogger.debug("Removing banned URL", "%s" % i)
                urlsList.remove(i)

        # Take the first URL from the list
        # randUrls = List.randomize( urlsList ) + failoverUrls

        sURL = urlsList[0]

        # If we have banned URLs, and several URLs at disposals, we make sure that the selected sURL
        # is not on a host which is banned. If it is, we take the next one in the list using __selectUrl
        # If we have banned URLs, and several URLs at disposals, we make sure that the selected sURL
        # is not on a host which is banned. If it is, we take the next one in the list using __selectUrl

        if len(self.__bannedUrls) > 0 and self.__nbOfUrls > 2:  # when we have multiple services then we can
            # have a situation when two services are running on the same machine with different ports...
            retVal = Network.splitURL(sURL)
            nexturl = None
            if retVal["OK"]:
                nexturl = retVal["Value"]

                found = False
                for i in self.__bannedUrls:
                    retVal = Network.splitURL(i)
                    if retVal["OK"]:
                        bannedurl = retVal["Value"]
                    else:
                        break
                    # We found a banned URL on the same host as the one we are running on
                    if nexturl[1] == bannedurl[1]:
                        found = True
                        break
                if found:
                    nexturl = self.__selectUrl(nexturl, urlsList[1:])
                    if nexturl:  # an url found which is in different host
                        sURL = nexturl
        gLogger.debug("Discovering URL for service", "%s -> %s" % (self._destinationSrv, sURL))
        return S_OK(sURL)

    def __selectUrl(self, notselect, urls):
        """In case when multiple services are running in the same host, a new url has to be in a different host
        Note: If we do not have different host we will use the selected url...

        :param notselect: URL that should NOT be selected
        :param list urls: list of potential URLs

        :return: str -- selected URL
        """
        url = None
        for i in urls:
            retVal = Network.splitURL(i)
            if retVal["OK"]:
                if retVal["Value"][1] != notselect[1]:  # the hosts are different
                    url = i
                    break
                else:
                    gLogger.error(retVal["Message"])
        return url

    def __checkThreadID(self):
        """
        ..warning:: just guessing....
        This seems to check that we are not creating a client and then using it
        in a multithreaded environment.
        However, it is triggered only if self.__enableThreadCheck is to True, but it is
        hardcoded to False, and does not seem to be modified anywhere in the code.
        """
        if not self.__initStatus["OK"]:
            return self.__initStatus
        cThID = thread.get_ident()
        if not self.__allowedThreadID:
            self.__allowedThreadID = cThID
        elif cThID != self.__allowedThreadID:
            msgTxt = """
=======DISET client thread safety error========================
Client %s
can only run on thread %s
and this is thread %s
===============================================================""" % (
                str(self),
                self.__allowedThreadID,
                cThID,
            )
            gLogger.error("DISET client thread safety error", msgTxt)
            # raise Exception( msgTxt )

    def _connect(self):
        """Establish the connection.
        It uses the URL discovered in __discoverURL.
        In case the connection cannot be established, __discoverURL
        is called again, and _connect calls itself.
        We stop after trying self.__nbOfRetry * self.__nbOfUrls

        :return: S_OK()/S_ERROR()
        """
        # Check if the useServerCertificate configuration changed
        # Note: I am not really sure that  all this block makes
        # any sense at all since all these variables are
        # evaluated in __discoverCredentialsToUse
        if gConfig.useServerCertificate() != self.__useCertificates:
            if self.__forceUseCertificates is None:
                self.__useCertificates = gConfig.useServerCertificate()
                self.kwargs[self.KW_USE_CERTIFICATES] = self.__useCertificates
                # The server certificate use context changed, rechecking the transport sanity
                result = self.__checkTransportSanity()
                if not result["OK"]:
                    return result

        # Take all the extra credentials
        self.__discoverExtraCredentials()
        if not self.__initStatus["OK"]:
            return self.__initStatus
        if self.__enableThreadCheck:
            self.__checkThreadID()

        gLogger.debug("Trying to connect to: %s" % self.serviceURL)
        try:
            # Calls the transport method of the apropriate protocol.
            # self.__URLTuple[1:3] = [server name, port, System/Component]
            transport = gProtocolDict[self.__URLTuple[0]]["transport"](self.__URLTuple[1:3], **self.kwargs)
            # the socket timeout is the default value which is 1.
            # later we increase to 5
            retVal = transport.initAsClient()
            # We try at most __nbOfRetry each URLs
            if not retVal["OK"]:
                gLogger.warn("Issue getting socket:", "%s : %s : %s" % (transport, self.__URLTuple, retVal["Message"]))
                # We try at most __nbOfRetry each URLs
                if self.__retry < self.__nbOfRetry * self.__nbOfUrls - 1:
                    # Recompose the URL (why not using self.serviceURL ? )
                    url = "%s://%s:%d/%s" % (
                        self.__URLTuple[0],
                        self.__URLTuple[1],
                        int(self.__URLTuple[2]),
                        self.__URLTuple[3],
                    )
                    # Add the url to the list of banned URLs if it is not already there. (Can it happen ? I don't think so)
                    if url not in self.__bannedUrls:
                        gLogger.warn("Non-responding URL temporarily banned", "%s" % url)
                        self.__bannedUrls += [url]
                    # Increment the retry counter
                    self.__retry += 1
                    # 16.07.20 CHRIS: I guess this setSocketTimeout does not behave as expected.
                    # If the initasClient did not work, we anyway re-enter the whole method,
                    # so a new transport object is created.
                    # However, it migh be that this timeout value was propagated down to the
                    # SocketInfoFactory singleton, and thus used, but that means that the timeout
                    # specified in parameter was then void.

                    # If it is our last attempt for each URL, we increase the timeout
                    if self.__retryCounter == self.__nbOfRetry - 1:
                        transport.setSocketTimeout(5)  # we increase the socket timeout in case the network is not good
                    gLogger.info("Retry connection", ": %d to %s" % (self.__retry, self.serviceURL))
                    # If we tried all the URL, we increase the global counter (__retryCounter), and sleep
                    if len(self.__bannedUrls) == self.__nbOfUrls:
                        self.__retryCounter += 1
                        # we run only one service! In that case we increase the retry delay.
                        self.__retryDelay = 3.0 / self.__nbOfUrls if self.__nbOfUrls > 1 else 2
                        gLogger.info("Waiting %f seconds before retry all service(s)" % self.__retryDelay)
                        time.sleep(self.__retryDelay)
                    # rediscover the URL
                    self.__discoverURL()
                    # try to reconnect
                    return self._connect()
                else:
                    return retVal
        except Exception as e:
            gLogger.exception(lException=True, lExcInfo=True)
            return S_ERROR("Can't connect to %s: %s" % (self.serviceURL, repr(e)))
        # We add the connection to the transport pool
        gLogger.debug("Connected to: %s" % self.serviceURL)
        trid = getGlobalTransportPool().add(transport)

        return S_OK((trid, transport))

    def _disconnect(self, trid):
        """Disconnect the connection.

        :param str trid: Transport ID in the transportPool
        """
        getGlobalTransportPool().close(trid)

    @staticmethod
    def _serializeStConnectionInfo(stConnectionInfo):
        """We want to send tuple but we need to convert
        into a list
        """
        serializedTuple = [list(x) if isinstance(x, tuple) else x for x in stConnectionInfo]

        return serializedTuple

    def _proposeAction(self, transport, action):
        """Proposes an action by sending a tuple containing

          * System/Component
          * Setup
          * VO
          * action
          * extraCredentials

        It is kind of a handshake.

        The server might ask for a delegation, in which case it is done here.
        The result of the delegation is then returned.

        :param transport: the Transport object returned by _connect
        :param action: tuple (<action type>, <action name>). It depends on the
                       subclasses of BaseClient. <action type> can be for example
                       'RPC' or 'FileTransfer'

        :return: whatever the server sent back

        """
        if not self.__initStatus["OK"]:
            return self.__initStatus
        stConnectionInfo = ((self.__URLTuple[3], self.setup, self.vo), action, self.__extraCredentials, DIRAC.version)

        # Send the connection info and get the answer back
        retVal = transport.sendData(S_OK(BaseClient._serializeStConnectionInfo(stConnectionInfo)))
        if not retVal["OK"]:
            return retVal
        serverReturn = transport.receiveData()

        # TODO: Check if delegation is required. This seems to be used only for the GatewayService
        if serverReturn["OK"] and "Value" in serverReturn and isinstance(serverReturn["Value"], dict):
            gLogger.debug("There is a server requirement")
            serverRequirements = serverReturn["Value"]
            if "delegate" in serverRequirements:
                gLogger.debug("A delegation is requested")
                serverReturn = self.__delegateCredentials(transport, serverRequirements["delegate"])
        return serverReturn

    def __delegateCredentials(self, transport, delegationRequest):
        """Perform a credential delegation. This seems to be used only for the GatewayService.
        It calls the delegation mechanism of the Transport class. Note that it is not used when
        delegating credentials to the ProxyDB

        :param transport: the Transport object returned by _connect
        :param delegationRequest: delegation request

        :return: S_OK()/S_ERROR()
        """
        retVal = gProtocolDict[self.__URLTuple[0]]["delegation"](delegationRequest, self.kwargs)
        if not retVal["OK"]:
            return retVal
        retVal = transport.sendData(retVal["Value"])
        if not retVal["OK"]:
            return retVal
        return transport.receiveData()

    def __checkTransportSanity(self):
        """Calls the sanity check of the underlying Transport object
        and stores the result in self.__idDict.
        It is checked at the creation of the BaseClient, and when connecting
        if the use of the certificate has changed.

        :return: S_OK()/S_ERROR()
        """
        if not self.__initStatus["OK"]:
            return self.__initStatus
        retVal = gProtocolDict[self.__URLTuple[0]]["sanity"](self.__URLTuple[1:3], self.kwargs)
        if not retVal["OK"]:
            return retVal
        idDict = retVal["Value"]
        for key in idDict:
            self.__idDict[key] = idDict[key]
        return S_OK()

    def __setKeepAliveLapse(self):
        """Select the maximum Keep alive lapse between
        150 seconds and what is specifind in kwargs[KW_KEEP_ALIVE_LAPSE],
        and sets it in kwargs[KW_KEEP_ALIVE_LAPSE]

        :return: S_OK()/S_ERROR()
        """
        kaa = 1
        if self.KW_KEEP_ALIVE_LAPSE in self.kwargs:
            try:
                kaa = max(0, int(self.kwargs[self.KW_KEEP_ALIVE_LAPSE]))
            except Exception:
                pass
        if kaa:
            kaa = max(150, kaa)
        self.kwargs[self.KW_KEEP_ALIVE_LAPSE] = kaa
        return S_OK()

    def _getBaseStub(self):
        """Returns a list with [self._destinationSrv, newKwargs]
        self._destinationSrv is what was given as first parameter of the init serviceName

        newKwargs is an updated copy of kwargs:
          * if set, we remove the useCertificates (KW_USE_CERTIFICATES) in newKwargs

        This method is just used to return information in case of error in the InnerRPCClient

        :return: tuple
        """
        newKwargs = dict(self.kwargs)
        # Remove useCertificates as the forwarder of the call will have to
        # independently decide whether to use their cert or not anyway.
        if "useCertificates" in newKwargs:
            del newKwargs["useCertificates"]
        return [self._destinationSrv, newKwargs]

    def __bool__(self):
        return True

    # For Python 2 compatibility
    __nonzero__ = __bool__

    def __str__(self):
        return "<DISET Client %s %s>" % (self.serviceURL, self.__extraCredentials)
